"""
.. module:: theories.classy

:Synopsis: Managing the CLASS cosmological code
:Author: Jesus Torrado, Benjamin Audren (CLASS import and ``get_cl``)

.. |br| raw:: html

   <br />

This module imports and manages the CLASS cosmological code.

.. note::

   **If you use this cosmological code, please cite it as:**
   |br|
   `D. Blas, J. Lesgourgues, T. Tram, "The Cosmic Linear Anisotropy Solving System (CLASS). Part II: Approximation schemes"
   (arXiv:1104.2933) <https://arxiv.org/abs/1104.2933>`_

.. note::

   CLASS is renamed ``classy`` for most purposes within cobaya, due to CLASS' name being
   a python keyword.

Usage
-----

If you are using a likelihood that requires some observable from CLASS,
simply add it to the theory block, specifying its ``path`` if you have not installed CLASS
globally. You can specify any parameter that CAMB or CLASS understand within the ``theory``
sub-block of the ``params`` block:

.. code-block:: yaml

   theory:
     classy:

   params:
     theory:
       [any param that CLASS understands, fixed, sampled or derived]

.. note::

   Some parameter names can be specified in CAMB nomenclature, such that the ``classy`` and
   ``camb`` theory blocks can be swapped easily. If there is a conflict, i.e. CLASS and
   CAMB give different parameters the same name, CAMB meaning takes priority.

   .. todo:: Not ideal. Working on it.


Installation
------------

.. note::

   If you already have your own version of CLASS, just make sure that the Python interface
   has been compiled, take note of its installation path and specify it using the
   ``path`` option in the ``classy`` block of the input.

CLASS' python interface utilises the ``cython`` compiler. If typing ``cython`` in the
shell produces an error, install it with ``pip install cython --user``.

.. note::
   The fast way, assuming you are installing all your cosmological codes under
   ``/path/to/cosmo/``:

   .. code:: bash

      $ cd /path/to/cosmo/
      $ git clone https://github.com/lesgourg/class_public.git
      $ mv class_public CLASS
      $ cd CLASS
      $ make

   If the **second** line produces an error (because you don't have ``git`` installed),
   try

   .. code:: bash

      $ cd /path/to/cosmo/
      $ wget https://github.com/lesgourg/class_public/archive/master.zip
      $ unzip master.zip
      $ rm master.zip
      $ mv class_public-master CLASS
      $ cd CLASS
      $ make

If the instructions above failed, follow those in the
`official CLASS web page <http://class-code.net/>`_ to get CLASS compiled with the Python
interface ready.

If you modify CLASS and add new variables, you don't need to let cobaya now, but make
sure that the variables you create are exposed in the Python
interface (contact CLASS' developers if you need help with that).

.. warning::

   At this moment, CLASS is not compatible with ``gcc`` version 5.0 and above
   (type ``gcc --version`` to check yours). This would cause an error when trying to use
   CLASS in cobaya, containing a line similar to

   .. code::

      /home/<username>/.local/lib/python2.7/site-packages/classy.so: undefined symbol: _ZGVbN2v_sin

   To solve it, open the file ``Makefile`` in the CLASS folder and add change the line
   stating ``CCFLAG = -g -fPIC`` to

   .. code:: make

      CCFLAG = -g -fPIC -fno-tree-vectorize

   Finally, clean and recompile, using ``make clean ; make``.

   **Source:** `CLASS issue #99 in github <https://github.com/lesgourg/class_public/issues/99>`_

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import sys
import os
import numpy as np
from copy import deepcopy

# Local
from cobaya.theory import Theory
from cobaya.log import HandledException
from cobaya.conventions import subfolders, _theory
from cobaya.tools import get_path_to_installation
from cobaya.install import user_flag_if_needed

# Logger
import logging
log = logging.getLogger(__name__)


# Dictionary of CAMB->CLASS names (only containing the different ones)
camb_to_classy = {"ombh2":  "omega_b",
                  "omegab": "Omega_b",
                  "omch2":  "omega_cdm",
                  "omegac": "Omega_cdm",
                  "omegav": "Omega_Lambda",
                  "omegam": "Omega_m",
                  "As": "A_s",
                  "ns": "n_s",
                  "tau": "tau_reio",
                  "zre": "z_reio",
}

classy_to_camb = dict([[v,k] for k,v in camb_to_classy.iteritems()])


class classy(Theory):
    def initialise(self):
        """Importing CLASS from the correct path, if given, and if not, globally."""
        # If path not given, try using general path to modules
        path_to_installation = get_path_to_installation()
        if not self.path and path_to_installation:
            self.path = os.path.join(
                path_to_installation, subfolders[_theory], "CLASS")
        if self.path:
            log.info("Importing *local* CLASS from "+self.path)
            if not os.path.exists(self.path):
                log.error("The given folder does not exist: '%s'", self.path)
                raise HandledException
            try:
                classy_path = ''
                classy_build_path = os.path.join(self.path, "python", "build")
                for elem in os.listdir(classy_build_path):
                    if elem.find("lib.") != -1:
                        classy_path = os.path.join(classy_build_path, elem)
                        break
                # Inserting the previously found path into the list of import folders
                sys.path.insert(0, classy_path)
                from classy import Class, CosmoSevereError, CosmoComputationError
            except OSError:
                log.error("Either CLASS is not in the given folder,\n"
                          "'%s',\n or you have not compiled it.", self.path)
                raise HandledException
        else:
            log.info("Importing *global* CLASS.")
            try:
                from classy import Class, CosmoSevereError, CosmoComputationError
            except ImportError:
                log.error(
                    "Couldn't find the CLASS python interface.\n"
                    "Make sure that you have compiled it, and that you either\n"
                    " (a) specify a path (you didn't) or\n"
                    " (b) install the Python interface globally with\n"
                    "     '/path/to/class/python/python setup.py install --user'")
                raise HandledException
        # Propate errors up
        global CosmoComputationError
        global CosmoSevereError
        # Generate states, to avoid recomputing
        self.n_states = 3
        self.states = [{"classy": Class(), "params": None, "last": 0, "derived": None}
                       for i in range(self.n_states)]
        # Fixed input
        # Default: no output -- see `needs` method.
        self.input_default = {"output": ""}
        # Precision (fixed at the theory block level)
        if self.precision:
            if "sBBN file" in self.precision:
                self.precision["sBBN file"] = os.path.join(self.path, self.precision["sBBN file"])
            self.input_default.update(self.precision)

    def current_state(self):
       lasts = [self.states[i]["last"] for i in range(self.n_states)]
       return self.states[lasts.index(max(lasts))]

    def set(self, params_values_dict, i_state):
        # Feed the arguments defining the cosmology to the cosmological code
        self.input_default.update(dict(
            [(camb_to_classy.get(p,p),v) for p,v in params_values_dict.items()]))
        # Generate and save
        log.debug("Setting parameters: %r", self.input_default)
        self.states[i_state]["params"] = deepcopy(self.input_default)
        self.states[i_state]["classy"].struct_cleanup()
        self.states[i_state]["classy"].set(**self.input_default)

    def compute(self, derived=None, **params_values_dict):
        lasts = [self.states[i]["last"] for i in range(self.n_states)]
        try:
            # are the parameter values there already?
            i_state = (i for i in range(self.n_states)
                       if self.states[i]["params"] == params_values_dict).next()
            # Get (pre-computed) derived parameters
            if derived == {}:
                derived.update(dict([[p,v] for p,v in
                                     zip(self.output_params, self.states[i_state]["derived"])]))
            log.debug("Re-using computed results (state %d)", i_state)
        except StopIteration:
            # update the (first) oldest one and compute
            i_state = lasts.index(min(lasts))
            log.debug("Computing (state %d)", i_state)
            self.set(params_values_dict, i_state)
            try:
                if "Cl" in self.input_default["output"]:
                    self.states[i_state]["classy"].compute(["lensing"])
            # "Valid" failure of CLASS: parameters too extreme -> log and report
            except CosmoComputationError:
                log.exception("CLASS computation failed. Assigning null likelihood. "
                              "See error information below.")
                return False
            # CLASS not correctly initialised, or input parameters not correct
            except CosmoSevereError:
                log.error("Error setting parameters or computig results -- "
                          "see CLASS's error trace below.\n")
                raise
            # Prepare derived parameters
            if derived == {}:
                derived.update(self.get_derived(i_state))
                # Careful: next step must keep the order
                self.states[i_state]["derived"] = [derived[p] for p in self.output_params]
        # make this one the current one by decreasing the antiquity of the rest
        for i in range(self.n_states):
            self.states[i]["last"] -= max(lasts)
        self.states[i_state]["last"] = 1
        return True

    def needs(self, arguments):
        for k,v in arguments.items():
            if k == "l_max":
                # Take the max of the requested ones
                self.input_default["l_max_scalars"] = max(v, self.input_default.get("l_max_scalars", 0))
            elif k == "Cl":
                if any([("T" in cl) for cl in v]):
                       self.input_default["output"] += " tCl"
                if any([(("E" in cl) or ("B" in cl)) for cl in v]):
                       self.input_default["output"] += " pCl"
                # For modern experiments, always lensed Cl's!
                self.input_default["output"] += " lCl"
                self.input_default["lensing"] = "yes"
            else:
                log.error("'%s' does not understand the requirement '%s:%s'.",
                    self.__class__.__name__,k,v)
                raise HandledException
        # Derived parameters (if some need some additional computations)
        if "sigma8" in self.output_params:
            self.input_default["output"] += " mPk"
            self.input_default["P_k_max_h/Mpc"] = max(1, self.input_default.get("P_k_max_h/Mpc", 0))
        # Cleanup
        self.input_default["output"] = " ".join(set(self.input_default["output"].split()))

    def get_derived(self, i_state):
        """
        Populates a dictionary of derived parameters with their values, using the
        state #`i_state`.
        """
        derived = {}
        classy = self.states[i_state]["classy"]
        p_classy = [camb_to_classy.get(p,p) for p in self.output_params]
        derived_aux = classy.get_current_derived_parameters(p_classy)
        derived.update(
            dict([(classy_to_camb.get(p,p),v) for p,v in derived_aux.iteritems()]))
        # If another method for getting derived parameters added in the future,
        # recover "if None in derived:" from the CAMB interface
        # Check if all processes (actually does nothing here)
        try:
            first_None = (p for p,v in derived.iteritems() if v==None).next()
            log.error("Derived param '%s' not working in the CLASS interface", p)
            raise HandledException
        except StopIteration:
            pass # all well!
        return derived

    def get_cl(self):
        """
        Returns the :math:`C_{\ell}` from the cosmological code in :math:`\mu {\\rm K}^2`
        """
        current_classy = self.current_state()["classy"]
        # get C_l^XX from the cosmological code
        cl = current_classy.lensed_cl(self.input_default["l_max_scalars"])
        # convert dimensionless C_l's to C_l in muK**2
        T = current_classy.T_cmb()
        for key in cl:
            # All quantities need to be multiplied by this factor, except the
            # phi-phi term, that is already dimensionless
            if key not in ['pp', 'ell']:
                cl[key] *= (T*1.e6)**2
        return cl


# Installation routines ###################################################################

def is_installed(**kwargs):
    try:
        import classy
    except:
        return False
    return True

def install(force=False, path=None ,**kwargs):
    log.info("Installing pre-requisites...")
    import pip
    exit_status = pip.main(["install", "cython", "--upgrade"] + user_flag_if_needed())
    if exit_status:
        log.error("Could not install pre-requisite: cython")
        return False
    log.info("Downloading...")
    parent_path = os.path.abspath(os.path.join(path, ".."))
    from wget import download, bar_thermometer
    try:
        filename = download("https://github.com/lesgourg/class_public/archive/v2.6.1.tar.gz",
                            out=parent_path, bar=bar_thermometer)
    except:
        log.error("Error downloading the latest release of CLASS.")
        return False
    print ""
    classy_path_decompressed = os.path.join(parent_path,
        os.path.splitext(os.path.splitext(os.path.basename(filename))[0])[0])
    classy_path = os.path.join(parent_path, "CLASS")
    if force and os.path.exists(classy_path):
        from shutil import rmtree
        rmtree(classy_path)
    log.info("Uncompressing...")
    import tarfile
    tar = tarfile.open(filename, "r:gz")
    try:
        tar.extractall(parent_path)
        tar.close()
        os.remove(filename)
    except:
        log.error("Error decompressing downloaded file! Corrupt file?)")
    os.rename(classy_path_decompressed, classy_path)
    from subprocess import Popen, PIPE
    working = False
    patch = False
    # Patch for gcc>=5
    while not working:
        os.chdir(classy_path)
        if patch:
            # patch (hopefully will be removed in the future)
            log.info("TRYING AGAIN: Patching for gcc>=5...")
            process_makeclean = Popen(["make", "clean"], stdout=PIPE, stderr=PIPE)
            out, err = process_makeclean.communicate()
            os.chdir("./python")
            process_pythonclean = Popen(["python", "setup.py", "clean"], stdout=PIPE, stderr=PIPE)
            out, err = process_pythonclean.communicate()
            os.chdir("..")
            with open(os.path.join(classy_path, "python/setup.py"), "r") as setup:
                lines = setup.readlines()
                i_offending = (i for i,l in enumerate(lines) if "libraries=" in l).next()
                lines[i_offending] = "libraries=['class', 'mvec', 'm'],\n"
            with open(os.path.join(classy_path, "python/setup.py"), "w") as setup:
                setup.write("".join(lines))
        log.info("Compiling...")
        process_make = Popen(["make"], stdout=PIPE, stderr=PIPE)
        out, err = process_make.communicate()
        if process_make.returncode:
            log.info(out)
            log.info(err)
            log.error("Compilation failed!")
            return False
        log.info("Installing python package...")
        os.chdir("./python")
        process_install = Popen(["python", "setup.py", "install"] + user_flag_if_needed(),
                                stdout=PIPE, stderr=PIPE)
        out, err = process_install.communicate()
        if process_install.returncode:
            log.info(out)
            log.info(err)
            log.error("Installation failed!")
            return False
        # If installed but not importable, patch and try again
        if not is_installed():
            if patch:
                break
            patch = True
        else:
            working = True
    return True



"""
if I want to sample from f(x) but the lik knows x, I need to 3 things:
1. x is the parameter known by the likelihood
2. x = exp(logx)
3. I must drop logx before passing the parameters to the likelihood.

New proposed syntax (update development notes!)

params:
  logx:
    prior: [...]
    drop:
  x: lambda logx: np.exp(x)

Now, loading params info consists of creating:
- A list of "input" parameters, which is exactly the same as the "sampled" ones:
  --> those that have a "prior"
- A list of "likelihood" parameters: those that will be passed tp the likelihood;
  Identified because they are not derived, i.e., are assigned a value (not a dict)
  or a dict that does not contain a "prior".
- A list of functions to translate between both.

OJO: pasar los fijos al inicializar, porque puede que cambie la forma de inicializar segun
el valor de los fijos!


CONFLICTO:

No esta claso, si escribo
a: f(c)
b: g(c)
cual es IN y cual es OUT:

Dos soluciones: separar bloques in y out en INPUT, o forzar OUT (derived) to be
a: f(c)
b:
   derived: g(c)

DECIR QUE NO SE PERMITE FIJAR UNO DEFINIDO POR FUNCION (NO PODRIA PONER EL "DROP"!)

# Assume that the *un*known function arguments are likelihood output parameters
In particular, args of COSMO derived parameters are assumed to be COSMO output parameters!

Una razon para usar solo un nivel de derived: si el parametro "intermedio" es tan importante
entonces lo propio es incorporarlo al codigo cosmologico!!!
(Se puede hacer 2 niveles de manera simple???)
DE TODAS MANERAS, LO IMPORTANTE ES QUE LO *PERMITE* HACER (SIN CAMBIAR CODIGO),
AUNQUE NO SEA LO MAS COMODO! (Si lo quieres mas sencillo, te las defines aparte)
PONER EJEMPLO DE SCRIPTED CALL PARA HACER COSAS LIGERAMENTE MAS COMPLICADAS.

Otra razon para no usar dos niveles: es mucho mas dificil reconocer a que likelihood
pertence cada *sampled*, si pueden ser funcion de otros *sampled* que no son input.

So, no, *just one level!!!*

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
from six import string_types

# Global
from collections import OrderedDict as odict
from numbers import Number
from inspect import getargspec
from itertools import chain
from ast import parse

# Local
from cobaya.conventions import _prior, _p_drop, _p_derived, _p_label, _theory
from cobaya.tools import get_external_function, ensure_nolatex
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)


def is_fixed_param(info_param):
    """
    Returns True if `info_param` is a number, a string or a function.
    """
    return (isinstance(info_param, Number) or isinstance(info_param, string_types)
            or callable(info_param))

def is_sampled_param(info_param):
    """
    Returns True if `info_param` has a `%s` field.
    """%_prior
    return _prior in (info_param if hasattr(info_param, "get") else {})

def is_derived_param(info_param):
    """
    Returns False if `info_param` is "fixed" or "sampled".
    """
    return not(is_fixed_param(info_param) or is_sampled_param(info_param))

class Parametrisation(object):
    """
    Class managing parametrisation. Translates parameter between sampler+prior and likelihood
    """
    def __init__(self, info_params):
        # First, we load the parameters, not caring about whether they are understood by any likelihood
        # `input` contains the parameters (expected to be) understood by the likelihood,
        #   with its fixed value, its fixing function, or None if their value is given
        #   directly by the sampler.
        self._input = odict()
        self._input_funcs = dict()
        self._input_args = dict()
        self._output = odict()
        self._sampled = odict()
        self._derived = odict()
        self._derived_funcs = dict()
        self._derived_args = dict()
        self._theory_params = info_params.get(_theory,{}).keys()
        self._theory_args = dict()
        info_params_flat = odict([(p,info_params[_theory][p]) for p in self._theory_params])
        info_params_flat.update(odict([(p,info_params[p]) for p in info_params if p!=_theory]))
        for p, info in info_params_flat.iteritems():
            if is_fixed_param(info):
                self._input[p] = info if isinstance(info, Number) else None
                if self._input[p] == None:
                    self._input_funcs[p] = get_external_function(info)
                    self._input_args[p] = getargspec(self._input_funcs[p]).args
            if is_sampled_param(info):
                self._sampled[p] = info
                if not _p_drop in info:
                    self._input[p] = None
            if is_derived_param(info):
                self._derived[p] = info
                if _p_derived in (info or {}):
                    self._derived_funcs[p] = get_external_function(info[_p_derived])
                    self._derived_args[p] = getargspec(self._derived_funcs[p]).args
                else:
                    self._output[p] = None
        # Check that the sampled and derived params are all valid python variable names
        def valid(name):
            try:
                parse("%s=None"%name)
                return True
            except SyntaxError:
                return False
        for p in chain(self.sampled_params(),self.derived_params()):
            if not valid(p):
                is_in = p in self.sampled_params()
                eg_in = "  p_prime:\n    prior: ...\n  %s: 'lambda p_prime: p_prime'\n"%p
                eg_out = "  p_prime: 'lambda %s: %s'\n"%(p,p)
                log.error("Parameter name '%s' is not a valid Python variable name "
                          "(it needs to start with a letter or '_').\n"
                          "If this is an %s parameter of a likelihood or theory, "
                          "whose name you cannot change,%s define an associated "
                          "%s one with a valid name 'p_prime' as: \n\n%s",
                          p, "input" if is_in else "output", "" if is_in else " remove it and",
                          "sampled" if is_in else "derived", eg_in if is_in else eg_out)
                raise HandledException
        # Assume that the *un*known function arguments are likelihood output parameters
        args = (set(chain(*self._input_args.values()))
                .union(chain(*self._derived_args.values())))
        for p in self._input.keys() + self._sampled.keys() + self._output.keys():
            if p in args:
                args.remove(p)
        self._output.update({p:None for p in args})
        # if argument of a theory parameter, assume it's a theory parameter
        args_theory = (
            set(chain(
                *[v for p,v in self._input_args.iteritems() if p in self._theory_params]))
            .union(chain(
                *[v for p,v in self._derived_args.iteritems() if p in self._theory_params])))
        self._theory_params = self._theory_params + [p for p in args_theory
                                                     if not p in self._theory_params]
        # Useful sets: directly-sampled input parameters and directly "output-ed" derived
        self._directly_sampled = [p for p in self._input if p in self._sampled]
        self._directly_output = [p for p in self._derived if p in self._output]
        # Useful mapping: input params that vary if each sampled is varied
        self._sampled_input_dependence = odict(
            [[s,[i for i in self._input if s in self._input_args.get(i, {})]]
             for s in self._sampled])
        # From here on, some error control.
        dropped_but_never_used = (
            set([p for p,v in self._sampled_input_dependence.items() if not v])
            .difference(set(self._directly_sampled)))
        if dropped_but_never_used:
            log.error("Parameters %r are sampled but not passed to the likelihood or "
                      "theory code, neither ever used as arguments for any parameters. "
                      "Check that you are not using the '%s' tag unintentionally.",
                      list(dropped_but_never_used), _p_drop)
            raise HandledException
        # input params depend on input and sampled only, never on output/derived
        bad_input_dependencies = set(chain(*self._input_args.values())).difference(
            set(self.input_params()).union(set(self.sampled_params())))
        if bad_input_dependencies:
            log.error("Input parameters defined as functions can only depend on other "
                      "input parameters that are not defined as functions. "
                      "In particular, an input parameter cannot depend on %r",
                      list(bad_input_dependencies))
            raise HandledException
        # derived depend of input and output, never of sampled which are not input
        bad_derived_dependencies = set(chain(*self._derived_args.values())).difference(
            set(self.input_params()).union(set(self.output_params())))
        if bad_derived_dependencies:
            log.error("Derived parameters can only depend on input and output parameters, "
                      "never on sampled parameters that have been defined as a function. "
                      "In particular, a derived parameter cannot depend on %r",
                      list(bad_derived_dependencies))
            raise HandledException

    def input_params(self):
        return self._input

    def output_params(self):
        return self._output

    def sampled_params(self):
        return self._sampled

    def derived_params(self):
        return self._derived

    def theory_params(self):
        return self._theory_params

    def sampled_input_dependence(self):
        return self._sampled_input_dependence

    def to_input(self, sampled_params_values):
        as_odict = odict(zip(self.sampled_params(), sampled_params_values))
        # Fill first directly sampled input parameters
        self.input_params().update({p:as_odict[p] for p in self._directly_sampled})
        # Then evaluate the functions
        for p in self._input_funcs:
            args = {p:self.input_params().get(p, as_odict.get(p, None))
                    for p in self._input_args[p]}
            self.input_params()[p] = self._input_funcs[p](**args)
        return self.input_params()

    def to_derived(self, output_params_values):
        as_odict = odict(zip(self.output_params(), output_params_values))
        # Fill first derived parameters which are direct output parameters
        self.derived_params().update({p:as_odict[p] for p in self._directly_output})
        # Then evaluate the functions
        for p in self._derived_funcs:
            args = {p:self.input_params().get(p, as_odict.get(p, None))
                    for p in self._derived_args[p]}
            self.derived_params()[p] = self._derived_funcs[p](**args)
        return self.derived_params().values()

    def labels(self):
        """
        Returns a dictionary of LaTeX labels of the sampled and derived parameters.
        
        Uses the parameter name of no label has been given.
        """
        get_label = lambda p,info: ensure_nolatex((info if info else {}).get(_p_label,p))
        return odict([[p,get_label(p,info)] for p, info in
                      list(self.sampled_params().items())+list(self.derived_params().items())])

    # Python magic for the "with" statement
    def __enter__(self):
        return self
    def __exit__(self, exception_type, exception_value, traceback):
        return



"""
.. module:: samplers.polychord

:Synopsis: PolyChord nested sampler (REFERENCES)
:Author: Will Handley, Mike Hobson and Anthony Lasenby (for PolyChord), Jesus Torrado (for the wrapper only)

.. |br| raw:: html

   <br />

.. note::
   **If you use this sampler, please cite it as:**
   |br|
   `W.J. Handley, M.P. Hobson, A.N. Lasenby, "PolyChord: nested sampling for cosmology"
   (arXiv:1502.01856) <https://arxiv.org/abs/1502.01856>`_
   |br|
   `W.J. Handley, M.P. Hobson, A.N. Lasenby, "PolyChord: next-generation nested sampling"
   (arXiv:1506.00171) <https://arxiv.org/abs/1506.00171>`_

``PolyChord`` is an advanced
`nested sampler <http://projecteuclid.org/euclid.ba/1340370944>`_,
that uses slice sampling to sample within the
nested isolikelihoods contours. The use of slice sampling instead of rejection sampling
makes ``PolyChord`` specially suited for high-dimensional parameter spaces, and allows for
exploiting speed hierarchies in the parameter space. Also, ``PolyChord`` can explore
multi-modal distributions efficiently.

``PolyChord`` is an *external* sampler, not installed by default (just a wrapper for it).
You need to install it yourself following the instructions below.

Usage
-----

To use ``PolyChord``, you just need to include the sampler block (assuming the directory
structure described in the :ref:`installation instructions <pc_installation>`.

.. code-block:: yaml

   sampler:
     polychord:
       path: /path/to/cosmo/PolyChord
       [optional PolyChord input parameters]


The parameters of interest to the user, if not defined in the ``polychord:`` block,
are left to their defaults for PolyChord 1.9 (see ``README`` file in ``PolyChord``'s
installation folder). If you wish to modify them, copy them from the ``defaults.yaml``
file (reproduced below).

The wrapper for ``PolyChord`` deals automatically with some of ``PolyChord``'s original
input parameters (e.g. the number of dimensions and of derived parameters, and the
parameters related to the output). The speed hierarchy, specified as described in 
:ref:`mcmc_speed_hierarchy`, is exploited by default.

.. warning::

   If you want to sample with ``PolyChord``, your priors need to be bounded. This is
   because ``PolyChord`` samples uniformly from a bounded *hypercube*, defined by a
   non-trivial transformation for general unbounded priors.

The main output is the Monte Carlo sample of sequentially discarded *live points*, saved
in the standard sample format together with the ``input.yaml`` and ``full.yaml``
files (see :doc:`output`). The raw ``PolyChord`` products are saved in a 
subfolder of the output folder
(determined by the option ``base_dir`` -- default: ``polychord_output``).


.. _pc_installation:

Installation
------------

You will need a MPI-wrapped Fortran compiler
(it is rarely worth running PolyChord without MPI capabilities in realistic scenarios).
You should have an MPI implementation installed if you followed
:ref:`the instructions to install mpi4py <install_mpi>`.
On top of that, you need the Fortran compiler (we recommend the GNU one) and the
*development* package of MPI. Use your system's package manager to install them
(``gfortran`` and ``libopenmpi-dev`` in Debian-based systems), or contact your local
IT service. If everything is correctly installed, you should be able to type ``mpif90``
in the shell and not get a ``Command not found`` error.

Download PolyChord version 1.9 from
`CCPForge <https://ccpforge.cse.rl.ac.uk/gf/project/polychord/>`_
(login required, free registration).
Uncompress the file you just downloaded in the folder where you are installing the
samplers, and follow the instructions in the ``README`` file.

.. note::
   The fast way, assuming that you are in a GNU/Linux system, that you are following the 
   directory structure described in :ref:`directory_structure`, and the PolyChord file
   downloaded is called ``PolyChord_v1.9.tar.gz``:

   .. code:: bash

      $ cd /path/to/cosmo
      $ tar xzvf PolyChord_v1.9.tar.gz
      $ rm PolyChord_v1.9.tar.gz
      $ cd PolyChord
      $ make PyPolyChord
      $ echo -e "export LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:$PWD/lib" >> ~/.bashrc
      $ source ~/.bashrc       

If PolyChord has been compiled with MPI, you **must** call cobaya with ``mpirun`` if you
want to use PolyChord as a sampler.

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
import sys
import numpy as np

# Local
from cobaya.sampler import Sampler
from cobaya.mpi import get_mpi_rank
from cobaya.collection import Collection
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)


# attributes that need to be broadcasted
to_broadcast = ["pc_args"]

class polychord(Sampler):
    def initialise(self):
        """Imports the PolyChord sampler and prepares its arguments."""
        if not get_mpi_rank(): # rank = 0 (MPI master) or None (no MPI)
            log.info("Initialising")
        # Importing PolyChord from the correct path
        if self.path:
            if not get_mpi_rank():
                log.info("Importing PolyChord from %s", self.path)
                if not os.path.exists(self.path):
                    log.error(
                    "The path you indicated for PolyChord does not exist: "+self.path)
                    raise HandledException
            sys.path.insert(0, self.path)
            from PyPolyChord import PyPolyChord
        else:
            # Currently, not installable as a python package! This will ALWAYS fail
            log.error("You need to specify PolyChord's path.")
            raise HandledException
            # log.info("Importing *global* PolyChord")
            # try:
            #     import PyPolyChord
            # except ImportError:
            #     log.error(
            #         "Couldn't find PolyChord's python interface.\n"
            #         "Make sure that you have compiled it, and that you either\n"
            #         " (a) specify a path (you didn't) or\n"
            #         " (b) install PolyChord's python interface globally with ... ???\n")
            #     raise HandledException
        self.pc = PyPolyChord
        # Prepare arguments - get just the PolyChord arguments
        self.pc_args = dict([(p, getattr(self,p)) for p in
                        ["nlive", "num_repeats", "do_clustering", "precision_criterion",
                         "max_ndead", "boost_posterior", "feedback", "update_files",
                         "posteriors", "equals", "cluster_posteriors", "write_resume",
                         "read_resume", "write_stats", "write_live", "write_dead",
                         "base_dir", "grade_frac", "grade_dims"]])
        # Ignore null-defined ones, so PolyChord sets them to the defaults
        self.pc_args = dict([(k,v) for k,v in self.pc_args.iteritems() if not v is None])
        # Fill the automatic ones
        if not "feedback" in self.pc_args:
            values = {logging.CRITICAL: 0, logging.ERROR: 0, logging.WARNING: 0,
                      logging.INFO: 1, logging.DEBUG: 2}
            self.pc_args["feedback"] = values[log.getEffectiveLevel()]
        self.pc_args["nDims"] = self.prior.d()
        self.pc_args["nDerived"] = (
            1 + len(self.likelihood.derived) + len(self.likelihood.names()))
        try:
            output_folder = getattr(self.output, "folder")
            output_prefix = getattr(self.output, "prefix")
        except AttributeError:
            # dummy output -- no resume!
            from tempfile import gettempdir
            output_folder = gettempdir()
            from random import random
            output_prefix = hex(int(random()*16**6))[2:]
            self.pc_args["read_resume"] = False
        self.pc_args["base_dir"] = os.path.join(output_folder, self.pc_args["base_dir"])
        self.pc_args["file_root"] = output_prefix
        if not get_mpi_rank():
            # Creating output folder, if it does not exist (just one process)
            if not os.path.exists(self.pc_args["base_dir"]):
                os.makedirs(self.pc_args["base_dir"])
            # Idem, a clusters folder if needed -- notice that PolyChord's default
            # is "True", here "None", hence the funny condition below
            if self.pc_args.get("do_clustering") is not False:
                try:
                    os.makedirs(os.path.join(self.pc_args["base_dir"], "clusters"))
                except OSError: # exists!
                    pass
            log.info("Storing raw PolyChord output in '%s'.", self.pc_args["base_dir"])
        # explotining the speed hierarchy
        # sort blocks by paramters order and check contiguity (required by PolyChord!!!)
        speeds, blocks = zip(*self.likelihood.speed_blocked_params(as_indices=True))
        speeds, blocks = np.array(speeds), np.array(blocks)
        # weird behaviour of np.argsort with there is only 1 block
        if len(blocks) > 1:
            sorting_indices = np.argsort(blocks, axis=0)
        else:
            sorting_indices = [0]
        speeds, blocks = speeds[sorting_indices], blocks[sorting_indices]
        if np.all([np.all(block==range(block[0], block[-1]+1)) for block in blocks]):
            log.warning("TODO: SPEED HIERARCHY EXPLOITATION DISABLED FOR NOW!!!")
#            self.pc_args["grade_frac"] = list(speeds)
#            self.pc_args["grade_dims"] = [len(block) for block in blocks]
#            log.info("Exploiting a speed hierarchy with speeds %r and blocks %r",
#                     speeds, blocks)
        else:
            log.warning("Some speed blocks are not contiguous: PolyChord cannot deal "
                        "with the speed hierarchy. Not exploting it.")
        # prior conversion from the hypercube
        limits = self.prior.limits()
        # Check if priors are bounded (nan's to inf)
        inf = np.where(np.isfinite(limits)==False)
        if len(inf[0]):
            params_names = self.prior.names()
            params = [params_names[i] for i in sorted(list(set(inf[0])))]
            log.error("PolyChord needs bounded priors, but the parameter(s) '"+
                      "', '".join(params)+"' is(are) unbounded.")
            raise HandledException
        locs = limits[:,0]
        scales = limits[:,1] - limits[:,0]
        self.pc_args["prior"] = lambda x: (locs + np.array(x)*scales).tolist()
        # We will need the volume of the prior domain, since PolyChord divides by it
        self.volume = np.prod(scales)
        # Done!
        if not get_mpi_rank():
            log.info("Calling PolyChord with arguments"+
                     "\n  ".join([""]+["%s : "%p+str(v) for p,v in self.pc_args.iteritems()]))

    def run(self):
        """
        Prepares the posterior function and calls ``PolyChord``'s ``run`` function.
        """
        # Prepare the posterior
        # Don't forget to multiply by the volume of the physical hypercube,
        # since PolyChord divides by it
        def logpost(p):
            logprior = self.prior.logp(p)
            logpost = logprior + np.log(self.volume)
            if logprior > -np.inf:
                derived = []
                logliks = self.likelihood.logps(p, derived=derived)
                logpost += sum(logliks)
            else:
                logliks = np.array([np.nan]*len(self.likelihoods))
            # derived parameters -- add physical ones!!!
            derived += [-logprior] + (-2*logliks).tolist()
            log.debug("logpost=%r, derived=%r", logpost, derived)
            return logpost, derived
        log.info("Sampling!")
        self.pc.run_nested_sampling(logpost, **self.pc_args)
        
    def close(self):
        """
        Loads the sample of live points from ``PolyChord``'s raw output and writes it
        (if ``txt`` output requested).
        """
        if not get_mpi_rank(): # process 0 or single (non-MPI process)
            log.info("Loading PolyChord's resuts: samples and evidences.")
            prefix = os.path.join(self.pc_args["base_dir"], self.pc_args["file_root"])
            sample = np.loadtxt(prefix+".txt")
            self.collection = Collection(self.prior, self.likelihood, self.output, name="1")
            n_sampled = len(self.prior.names())
            n_derived = len(self.likelihood.derived)
            n_liks = len(self.likelihood.names())
            for row in sample:
                self.collection.add(row[2:2+n_sampled],
                                    derived=row[2+n_sampled:2+n_sampled+n_derived+1],
                                    weight=row[0], logpost=-row[1],
                                    logprior=-row[-(1+n_liks)],
                                    logliks=[-0.5*lik for lik in row[-n_liks:]])
            # make sure that the points are written
            self.collection.out_update()
            # Prepare the evidence
            with open(prefix+".stats", "r") as statsfile:
                line = ""
                while not "Global evidence:" in line:
                    line = statsfile.readline()
                while not "log(Z)" in line:
                    line = statsfile.readline()
                self.logZ, self.logZstd = [
                    float(n) for n in line.split("=")[-1].split("+/-")]
# THE RESULTS CANNOT BE BROADCASTED BECAUSE POLYCHORD KILLS MPI!
#        # Broadcast results
#        if get_mpi():
#            bcast_from_0 = lambda attrname: setattr(self, attrname,
#                get_mpi_comm().bcast(getattr(self, attrname, None), root=0))
#            map(bcast_from_0, ["collection", "logZ", "logZstd"])
        if not get_mpi_rank(): # process 0 or single (non-MPI process)
            log.info("Finished! "+
                     "Raw PolyChord output stored in '%s'.", self.pc_args["base_dir"])

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.

        Returns:
           The sample ``Collection`` containing the sequentially discarded live points.
        """
        if not get_mpi_rank():
            return {"sample": self.collection, "logZ": self.logZ, "logZstd": self.logZstd}



"""
.. module:: likelihoods.gaussian

:Synopsis: Gaussian mock likelihood, for tests
:Author: Jesus Torrado

Gaussian *mock* likelihoods, aimed for tests.

The mean and covariance matrix for one or more modes must be specified with the options 
``mean`` and ``covmat`` respectively. The dimensionality of the likelihood and the number
of modes are guessed from this options (if they are consistent to each other).

The pdf is normalised to 1 when integrated over an infinite domain,
regardless of the number of modes.

The following example defines 3 modes in 2 dimensions, and expects two parameters whose
names start with ``test_`` and must be defined in the ``params`` block:

.. code-block:: yaml

   likelihood:
     gaussian:
       mock_prefix: test_
       mean: [ [0.1,0.1],
               [0.3,0.3],
               [0.4,0.5] ]
       cov:  [ [[0.01, 0],
                [0,    0.05]],
               [[0.02,  0.003],
                [0.003, 0.01]],
               [[0.01, 0],
                [0,    0.01]] ]


Derived parameters can be tracked, as many as sampled parameters times the number of modes,
and they represent the standarised parameters of each of the modes, i.e. those distributed
as :math:`\mathcal{N}(0,I)` around each mode (notice that if a mode is close to the
boundary of the prior, you should not expect to recoved a unit covariance matrix from the
sample).

A delay (in seconds) in the likelihood evaluation can be specified with the keyword
``delay``.

.. note::

   This module also provides functions to generate random means and covariances
   -- see automatic documentation below.

Options and defaults
--------------------

Simply copy this block in your input ``yaml`` file and modify whatever options you want
(you can delete the rest).

.. literalinclude:: ../cobaya/likelihoods/gaussian/defaults.yaml
   :language: yaml

"""

# Python 2/3 compatibility
from __future__ import division

# Global
import numpy as np
from scipy.stats import multivariate_normal, uniform, random_correlation
from cobaya.mpi import get_mpi_size

# Local
from cobaya.likelihood import Likelihood
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)


class gaussian(Likelihood):
    """
    Gaussian *mock* likelihood.
    """
    def initialise(self):
        """
        Initialises the gaussian distributions.
        """
        log.info("Initialising")
        # Load mean and cov, and check consistency of n_modes and dimensionality
        if self.mean is not None and self.cov is not None:
            # Wrap them in the right arrays (n_mode, param) and check consistency
            self.mean = np.atleast_1d(self.mean)
            while len(np.array(self.mean).shape) < 2:
                self.mean = np.array([self.mean])
            mean_n_modes, mean_dim = self.mean.shape
            self.cov = np.atleast_1d(self.cov)
            while len(np.array(self.cov).shape) < 3:
                self.cov = np.array([self.cov])
            cov_n_modes, cov_dim, cov_dim_2 = self.cov.shape
            if cov_dim != cov_dim_2:
                log.error("The covariance matrix(/ces) do not appear to be square!\n"
                          "Got %r", self.cov)
                raise HandledException
            if mean_dim != cov_dim:
                log.error(
                    "The dimensionalities guessed from mean(s) and cov(s) do not match!")
                raise HandledException
            if mean_n_modes != cov_n_modes:
                log.error(
                    "The numbers of modes guessed from mean(s) and cov(s) do not match!")
                raise HandledException
            if mean_dim != self.d():
                log.error(
                    "The dimensionality is %d (guessed from given means and covmats) "
                    "but was given %d mock parameters instead. " +
                    ("Maybe you forgot to specify the prefix by which to identify them?"
                     if self.mock_prefix else ""), mean_dim, len(self.input_params))
                raise HandledException
            self.n_modes = mean_n_modes
            if len(self.output_params) != self.d()*self.n_modes:
                log.error(
                    "The number of derived parameters must be equal to the dimensionality"
                    " times the number of modes, i.e. %d x %d = %d, but was given %d "
                    "derived parameters.", self.d(), self.n_modes, self.d()*self.n_modes,
                    len(self.output_params))
                raise HandledException
        else:
            log.error("You must specify both a mean (or a list of them) and a "
                      "covariance matrix, or a list of them.")
            raise HandledException
        self.gaussians = [multivariate_normal(mean=mean, cov=cov)
                          for mean, cov in zip(self.mean, self.cov)]
        # Prepare the transformation(s) for the derived parameters
        self.choleskyL = [np.linalg.cholesky(cov) for cov in self.cov]

    def logp(self, **params_values):
        """
        Computes the log-likelihood for a given set of parameters.
        """
        self.wait()
        # Prepare the vector of sampled parameter values
        x = np.array([params_values[p] for p in self.input_params])
        # Fill the derived parameters
        derived = params_values.get("derived")
        if derived is not None:
            for i in range(self.n_modes):
                standard = np.linalg.inv(self.choleskyL[i]).dot((x-self.mean[i]))
                derived.update(dict(
                    [(p,v) for p,v in
                     zip(self.output_params.keys()[i*self.d():(i+1)*self.d()],standard)]))
        # Compute the likelihood and return
        return (-np.log(self.n_modes) +
                 np.log(sum([gauss.pdf(x) for gauss in self.gaussians])))


# Scripts to generate random means and covariances #######################################

def random_mean(ranges, n_modes=1):
    """
    Returns a uniformly sampled point (as an array) within a list of limits `ranges`.

    The output of this function can be used directly as the value of the option `mean` of
    the :class:`likelihoods.gaussian`.

    If `n_modes>1`, returns an array of such points.
    """
    if get_mpi_size():
        print ("WARNING! "
               "Using with MPI: different process will produce different random results.")
    mean = np.array([uniform.rvs(loc=r[0], scale=r[1]-r[0], size=n_modes)
                     for r in ranges])
    mean = mean.T
    if n_modes == 1:
        mean = mean[0]
    return mean


def random_cov(ranges, O_std_min=1e-2, O_std_max=1, n_modes=1):
    """
    Returns a random covariance matrix, with standard deviations sampled log-uniformly
    from the lenght of the parameter ranges times `O_std_min` and `O_std_max`, and
    uniformly sampled correlation coefficients betweem `rho_min` and `rho_max`.

    The output of this function can be used directly as the value of the option `cov` of
    the :class:`likelihoods.gaussian`.

    If `n_modes>1`, returns a list of such matrices.
    """
    if get_mpi_size():
        print ("WARNING! "
               "Using with MPI: different process will produce different random results.")
    dim = len(ranges)
    scales = np.array([r[1]-r[0] for r in ranges])
    cov = []
    for i in range(n_modes):
        stds = scales * 10**(uniform.rvs(size=dim, loc=np.log10(O_std_min),
                                         scale=np.log10(O_std_max/O_std_min)))
        this_cov = np.diag(stds).dot(
            (random_correlation.rvs(dim*stds/sum(stds)) if dim > 1 else np.eye(1))
            .dot(np.diag(stds)))
        # Symmetrise (numerical noise is usually introduced in the last step)
        cov += [(this_cov+this_cov.T)/2]
    if n_modes == 1:
        cov = cov[0]
    return cov



"""
.. module:: samplers.evaluate

:Synopsis: Dummy "sampler": simply evaluates the likelihood.
:Author: Jesus Torrado

This is a *dummy* sampler that just evaluates the likelihood in a given (or sampled)
reference point.

You can use it to test your likelihoods.

To use it, simply make the ``sampler`` block:

.. code-block:: yaml

   sampler:
     evaluate:

If you want to fix the evaluation point, give just a value as a *reference*
for the parameters in the input file. For example:

.. code-block:: yaml

   params:
     a:
       prior:
         min: -1
         max:  1
       ref: 0.5
     b:
       prior:
         min: -1
         max:  1
       ref:
         dist: norm
         loc: 0
         scale: 0.1
     c:
       prior:
         min: -1
         max:  1

In this case, the posterior will be evaluated at ``a=0.5``, ``b`` sampled from a normal
pdf centred at ``0`` with standard deviation ``0.1``, and ``c`` will be sampled uniformly
between ``-1`` and ``1``.

"""

# Python 2/3 compatibility
from __future__ import division

# Local
from cobaya.sampler import Sampler
from cobaya.collection import Collection

# Logger
import logging
log = logging.getLogger(__name__)


class evaluate(Sampler):

    def initialise(self):
        """
        Creates a 1-point collection to store the point
        at which the posterior is evaluated.
        """ 
        self.one_point = Collection(self.parametrisation, self.likelihood,
                                    self.output, initial_size=1, name="1")
        log.info("Initialised!")
        
    def run(self):
        """
        First gets a reference point. If a single reference point is not given,
        the point is sampled from the reference pdf, and if that one is not defined either,
        the point is sampled from the prior.

        Then it evaluates the prior and likelihood(s) and stores them in the one-member 
        sample collection.
        """
        reference_point = self.prior.reference()
        log.info("Reference point:\n   "+
               "\n   ".join(["%s = %g"%(p,reference_point[i])
                             for i,p in enumerate(self.parametrisation.sampled_params())]))
        log.info("Evaluating prior and likelihoods...")
        logpost, logprior, logliks, derived = self.logposterior(reference_point)
        self.one_point.add(reference_point, derived=derived,
                           logpost=logpost, logprior=logprior, logliks=logliks)
        log.info("log-posterior  = %g", logpost)
        log.info("log-prior      = %g", logprior)
        log.info("log-likelihood = %g", sum(logliks))
        log.info("   "+"\n   ".join(["chi2_"+name+" = %g"%(-2*logliks[i])
                  for i,name in enumerate(self.likelihood)]))
        log.info("Done!")

    def close(self):
        """
        Writes the output: the point and its prior, posterior and likelihood.
        """
        self.one_point.out_update()

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.

        Returns:
           The sample ``Collection`` containing the sequentially discarded live points.
        """
        return {"sample": self.one_point}



"""
.. module:: output

:Synopsis: Generic output class and output drivers
:Author: Jesus Torrado

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
from copy import deepcopy
import datetime
        
# Local
from cobaya.yaml_custom import yaml_dump
from cobaya.conventions import _likelihood, _theory, _prior, _params
from cobaya.conventions import _sampler, _input_suffix, _full_suffix
from cobaya.conventions import separator
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)

class Output():
    def __init__(self, info):
        output_prefix = str(info["output_prefix"])
        if os.sep in output_prefix:
            self.folder = os.sep.join(output_prefix.split(os.sep)[:-1])
            self.prefix = output_prefix.split(os.sep)[-1]
            log.debug("Creating output folder '%s'", self.folder)
            # Notice that the folder cannot exist already.
            # If we are continuing a sample, that is dealt with by the invocation script.
            try:
                os.makedirs(self.folder)
            except OSError:
                log.error("Chain continuation not implemented. "
                          "If testing, delete the chain folder '%s' before invoking again.",
                          self.folder)
                raise HandledException
                # The exception that should be raised when this is implemented.
                raise OSError("Cannot create folder '%s'. Check your 'output_prefix'.", self.folder)
        else:
            self.folder = "."
            # safeguard against calling from chain folder
            if output_prefix == ".":
                output_prefix = ""
            self.prefix = output_prefix
        log.info("The output folder is '%s', and the output prefix is '%s'", 
                 self.folder, self.prefix)
        # Save the updated name and output_prefix: now relative to output folder
        self.info_input = deepcopy(info)
        self.info_input["output_prefix"] = self.prefix if self.prefix else "."
        # Output kind and collection extension
        self.kind = "txt"
        self.ext  = "txt"

    def set_full_info(self, full_info):
        self._full_info = full_info

    def dump_info(self, likelihood=None, theory=None, sampler=None, params=None):
        """
        Saves the info in the chain folder twice:
           - the input info.
           - idem, populated with the modules' defaults.
        """
        file_prefix = self.prefix
        if self.prefix:
            file_prefix += separator
        file_input = os.path.join(self.folder, file_prefix+_input_suffix+".yaml")
        file_full = os.path.join(self.folder, file_prefix+_full_suffix+".yaml")
        for f, info in [(file_input, self.info_input), (file_full, self._full_info)]:
            if os.path.isfile(f):
                log.error("Chain continuation not implemented. "
                          "If testing, delete the relevant chain folder.")
                raise HandledException
                # This is the error we should be raising when chain continuation is implem.
                raise OSError("Dumping the info: The file '%s' exists!", f)
            with open(f, "w") as f_out:
                f_out.write(yaml_dump(info, default_flow_style=False))

    def prepare_collection(self, name=None):
        if not name:
            name = (datetime.datetime.now().isoformat()
                    .replace("T","_").replace(":","-").replace(".","-"))
        file_name = os.path.join(
            self.folder, self.prefix+("_" if self.prefix else "")+name+"."+self.ext)
        return file_name, self.kind


class Output_MPI(Output):
    """
    MPI wrapper around the Output class.
    """
    def __init__(self, info):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        to_broadcast = ("folder", "prefix", "kind", "ext", "info_input")
        if rank == 0:
            Output.__init__(self, info)
        else:
            for var in to_broadcast:
                setattr(self, var, None)
        for var in to_broadcast:
            setattr(self, var, comm.bcast(getattr(self, var), root=0))

    def dump_info(self, *args, **kwargs):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank == 0:
            Output.dump_info(self, *args, **kwargs)


class Output_dummy(Output):
    """
    Dummy output class. Does nothing. Evaluates to 'False' as a class.
    """
    def __init__(self, info):
        log.info("No output requested. Doing nothing (or returning in scripted call).")
        self.info_input = deepcopy(info)
        # override all methods
        exclude = ["__nonzero__", "nullfunc", "update_info", "updated_info"]
        for attrname,attr in Output.__dict__.items():
            func_name = getattr(attr, "func_name", None)
            if func_name and not func_name in exclude:
                setattr(self, attrname, self.nullfunc)

    def nullfunc(self, *args, **kwargs):
        pass

    def __nonzero__(self):
        return False



"""
.. module:: input

:Synopsis: Input-related functions
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
from collections import OrderedDict as odict
from copy import deepcopy
from importlib import import_module

# Local
from cobaya.conventions import package, _defaults_file, _params, _p_label
from cobaya.conventions import _prior, _theory, _likelihood, _sampler, _external
from cobaya.tools import get_folder
from cobaya.yaml_custom import yaml_load_file
from cobaya.log import HandledException
from cobaya.parametrisation import is_sampled_param, is_derived_param

# Logger
import logging
log = logging.getLogger(__name__)


def load_input(input_file):
    """
    Loads general info, and splits it into the right parts.
    """
    file_name, extension = os.path.splitext(input_file)
    file_name = os.path.basename(file_name)
    if extension in (".yaml",".yml"):
        info = yaml_load_file(input_file)
        # if output_prefix not defined, default to input_file name (sans ext.) as prefix;
        if "output_prefix" not in info:
            info["output_prefix"] = file_name
        # warn if no output, since we are in shell-invocation mode.
        elif info["output_prefix"] is None:
            log.warning("WARNING: Output explicitly supressed with 'ouput_prefix: null'")
    else:
        log.error("Extension of input file '%s' not recognised.", input_file)
        raise HandledException
    return info


# MPI wrapper for loading the input info
def load_input_MPI(input_file):
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    # Load input (only one process does read the input file)
    if rank == 0:
        info = load_input(input_file)
    else:
        info = None
    info = comm.bcast(info, root=0)
    return info


def get_modules(*infos):
    """Returns modules all requested as an odict ``{kind: set([modules])}``.
    Priors are not included."""
    modules = odict()
    for info in infos:
        for field in [_theory, _likelihood, _sampler]:
            if field not in modules:
                modules[field] = set()
            modules[field] |= (lambda v: set(v) if v else set())(info.get(field))
            modules[field] = modules[field].difference(set([None]))
    # pop empty fields
    for k,v in modules.iteritems():
        if not v:
            modules.pop(k)
    return modules


def get_full_info(info):
    """
    Creates an updated info starting from the defaults for each module and updating it
    with the input info.
    """
    # Don't modify the original input!
    input_info = deepcopy(info)
    # Creates an equivalent info using only the defaults
    full_info = odict()
    default_params_info = odict()
    default_prior_info = odict()
    modules = get_modules(input_info)
    for block in modules:
        full_info[block] = odict()
        for module in modules[block]:
            # Start with the default class options
            full_info[block][module] = deepcopy(getattr(
                import_module(package+"."+block, package=package), "class_options", {}))
            # Go on with defaults
            path_to_defaults = os.path.join(get_folder(module, block), _defaults_file)
            try:
                default_module_info = yaml_load_file(path_to_defaults)
            except IOError:
                # probably an external module
                default_module_info = {block: {module: {}}}
                log.debug("Module %s:%s does not have a defaults file. "%(block, module) +
                          "Maybe it is an external module.")
            try:
                full_info[block][module].update(default_module_info[block][module] or {})
            except KeyError:
                log.error("The defaults file for '%s' should be structured "
                          "as %s:%s:{[options]}.", module, block, module)
                raise HandledException
            # Update the options with the input file
            # Consistency is checked only up to first level! (i.e. subkeys may not match)
            # First deal with cases "no options" and "external function"
            try:
                input_info[block][module] = input_info[block][module] or {}
            except TypeError:
                log.error("Your input info is not well formatted at the '%s' block. "
                          "It must be a dictionary {'%s':{options}, ...}. ", block, block)
                raise HandledException
            if not hasattr(input_info[block][module], "get"):
                input_info[block][module] = {_external: input_info[block][module]}
            options_not_recognised = (set(input_info[block][module])
                                      .difference(set([_external]))
                                      .difference(set(full_info[block][module])))
            if options_not_recognised:
                log.error("'%s' does not recognise some options: '%r'. "
                          "To see the allowed options, check out the file '%s'",
                          module, tuple(options_not_recognised), path_to_defaults)
                raise HandledException
            full_info[block][module].update(input_info[block][module])
            # Store default parameters and priors of class, and save to combine later
            if block == _likelihood:
                params_info = default_module_info.get(_params, {})
                full_info[block][module].update({_params:params_info})
                default_params_info[module] = params_info
                default_prior_info[module] = default_module_info.get(_prior, {})
    # Add priors info, after the necessary checks
    if _prior in input_info or any(default_prior_info.values()):
        full_info[_prior] = input_info.get(_prior, odict())
    for prior_info in default_prior_info.values():
        for name, prior in prior_info.iteritems():
            if full_info[_prior].get(name, prior) != prior:
                log.error("Two different priors cannot have the same name: '%s'.", name)
                raise HandledException
            full_info[_prior][name] = prior
    # Add parameters info, after the necessary updates and checks
    full_info[_params] = merge_params_info(input_info.get(_params, {}),
                                           defaults=default_params_info)
    # Rest of the options
    for k,v in input_info.iteritems():
        if k not in full_info:
            full_info[k] = v
    return full_info


def merge_params_info(params_info, defaults=None):
    """
    Merges input and default parameters info, after performing some consistency checks.
    """
    # First, merge defaults. Impose multiple defined (=shared) parameters have equal info
    defaults_merged = odict()
    for lik, params in defaults.iteritems():
        for p, info in params.iteritems():
            # if already there, check consistency
            if p in defaults_merged:
                log.debug("Parameter '%s' multiply defined.", p)
                if info != defaults_merged[p]:
                    log.error("Parameter '%s' multiply defined, but inconsistent info: "
                              "For likelihood '%s' is '%r', but for some other likelihood"
                              " it was '%r'. Check your defaults!",
                              p, lik, info, defaults_merged[p])
                    raise HandledException
            defaults_merged[p] = info
    # Combine with the input parameter info (make sure the theory parameters come first)
    params_info_copy = deepcopy(params_info)
    theory_params = params_info_copy.pop(_theory, None)
    info_updated = odict()
    if theory_params:
        info_updated.update({_theory: theory_params})
    info_updated.update(defaults_merged)
    info_updated.update(params_info)
    # Inherit labels (for sampled and derived) and min/max (just for derived params)
    getter = lambda info, key: getattr(info, "get", lambda x: None)(key)
    for p in defaults_merged:
        default_label = getter(defaults_merged[p], _p_label)
        if (default_label and
                (is_sampled_param(info_updated[p]) or is_derived_param(info_updated[p]))):
            info_updated[p][_p_label] = info_updated[p].get(_p_label) or default_label
        limits = ["min", "max"]
        default_limits = odict([[lim,getter(defaults_merged[p], lim)] for lim in limits])
        if default_limits.values() != [None, None] and is_derived_param(info_updated[p]):
            info_updated[p].update(default_limits)
    return info_updated



from cobaya.likelihoods.planck_clik_prototype import planck_clik_prototype

class planck_2015_lowl(planck_clik_prototype):
    pass

from cobaya.likelihoods.planck_clik_prototype import install as install_common
from cobaya.likelihoods.planck_clik_prototype import is_installed as is_installed_common

def is_installed(**kwargs):
    kwargs["name"] = __name__.split(".")[-1]
    return is_installed_common(**kwargs)

def install(**kwargs):
    kwargs["name"] = __name__.split(".")[-1]
    return install_common(**kwargs)



"""
.. module:: log

:Synopsis: Manages logging and error handling
:Author: Jesus Torrado

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import sys
import logging
import traceback

# Local
from cobaya.conventions import _debug, _debug_file
from cobaya.mpi import get_mpi, get_mpi_rank


class HandledException(Exception):
    """
    Dummy exception, to be raised when the originating exception
    has been cleanly handled and logged.
    """


def exception_handler(exception_type, value, trace_back):
    # Do nothing (just exit) if the exception has been handled and logged
    if exception_type == HandledException:
        return
    log = logging.getLogger("exception handler")
    line = "------------------------------------------------\n"
    log.critical(line[6:] + "\n" +
                 "".join(traceback.format_exception(exception_type, value, trace_back)) +
                 line)
    if exception_type == KeyboardInterrupt:
        log.critical("Interrupted by the user.")
        return
    log.critical(
        "Some unexpected ERROR ocurred. You can see the exception information above.\n"
        "We recommend trying to reproduce this error with '%s:True' in the input.\n"
        "If you cannot solve it yourself and need to report it, include the debug ouput,"
        "\nwhich you can send it to a file setting '%s:[some_file_name]'.",
        _debug, _debug_file)

    
def logger_setup(debug=None, debug_file=None):
    """
    Configuring the root logger, for its childs to inherit level, format and handlers.

    Level: if debug=True, take DEBUG. If numerical, use "logging"'s corresponding level.
    Default: INFO
    """
    if debug is True:
        level = logging.DEBUG
    elif debug in (False, None):
        level = logging.INFO
    else:
        level = int(debug)
    # Set the default level, to make sure the handlers have a higher one
    logging.root.setLevel(level)
    # Custom formatter
    class MyFormatter(logging.Formatter):
        def format(self, record):
            self._fmt = (
                ("[%d]"%get_mpi_rank() if get_mpi() else "") +
                "[%(module)s] " +
                {logging.ERROR: "*ERROR* ",
                 logging.WARNING: "*WARNING* "}.get(record.levelno, "") +
                "%(message)s")
            return logging.Formatter.format(self, record)
    # Configure stdout handler
    handle_stdout = logging.StreamHandler(sys.stdout)
    handle_stdout.setLevel(level)
    handle_stdout.setFormatter(MyFormatter())
    # log file? Create and reduce stdout level to INFO
    if debug_file:
        file_stdout = logging.FileHandler(debug_file, mode="w")
        file_stdout.setLevel(level)
        handle_stdout.setLevel(logging.INFO)
        file_stdout.setFormatter(MyFormatter())
        logging.root.addHandler(file_stdout)
    # Add stdout handler only once!
    if not any(h.stream == sys.stdout for h in logging.root.handlers):
        logging.root.addHandler(handle_stdout)
    # Configure the logger to manage exceptions
    sys.excepthook = exception_handler



"""
.. module:: collection

:Synopsis: Class to store the Montecarlo sample
:Author: Jesus Torrado

Class keeping track of the samples.

Basically, a wrapper around a `pandas.DataFrame`.

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
import numpy as np
import pandas as pd
from getdist import MCSamples

# Local
from cobaya.conventions import _weight, _chi2, _minuslogpost, _minuslogprior, _derived_pre
from cobaya.conventions import separator
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)


# Default chunk size for enlarging collections more efficiently
# If a factor is defined (as a fraction !=0), it is used instead
# (e.g. 0.10 to grow the number of rows by 10%)
enlargement_size = 100
enlargement_factor = None

class Collection():

    def __init__(self, parametrisation, likelihood, output=None,
                 initial_size=enlargement_size, name=None):
        self.name = name
        self.sampled_params = parametrisation.sampled_params().keys()
        self.derived_params = parametrisation.derived_params().keys()
        # Create the dataframe structure
        columns = [_weight, _minuslogpost]
        columns += list(self.sampled_params)
        columns += [_derived_pre+name for name in self.derived_params]
        columns += [_minuslogprior]
        self.chi2_names = [_chi2+separator+name for name in likelihood]
        columns += [_chi2] + self.chi2_names
        # Create the main data frame and the tracking index
        self.data = pd.DataFrame(np.zeros((initial_size,len(columns))), columns=columns)
        self._n = 0
        # Prepare labels and ranges for interfacing with GetDist
        self.labels = parametrisation.labels()


        # RANGES HERE!!!!
        

        # OUTPUT: Driver, file_name, index to track last sample written
        if output:
            self._n_last_out = 0
            self.file_name, self.driver = output.prepare_collection(name=self.name)
            self.out_prepare()
        else:
            self.driver = "dummy"

    def add(self, values, derived=None,
                  weight=1, logpost=None, logprior=None, logliks=None):
        self.enlarge_if_needed()
        self.data[_weight][self._n] = weight
        if logpost == None:
            try:
                logpost = logprior + sum(logliks)
            except ValueError:
                log.error("If a log-posterior is not specified, you need to pass "
                          "a log-likelihood and a log-prior.")
                raise HandledException
        self.data[_minuslogpost][self._n] = -logpost
        if logprior != None:
            self.data[_minuslogprior][self._n] = -logprior
        if logliks is not None:
            for name, value in zip(self.chi2_names, logliks):
                self.data[name][self._n] = -2*value
            self.data[_chi2][self._n] = sum(self.data[self.chi2_names].values[self._n])
        for name, value in zip(self.sampled_params, values):
            self.data[name][self._n] = value
        if derived is not None:
            for name, value in zip(self.derived_params, derived):
                self.data[_derived_pre+name][self._n] = value
        self._n += 1

    def enlarge_if_needed(self):
        if self._n >= self.data.shape[0]:
            if enlargement_factor:
                enlarge_by = self.data.shape[0]*enlargement_factor
            else:
                enlarge_by = enlargement_size
            self.data = pd.concat([self.data,
                pd.DataFrame(np.zeros((enlarge_by, self.data.shape[1])), 
                             columns=self.data.columns,
                             index=np.arange(self.n(),self.n()+enlarge_by))])
        
    # Retrieve-like methods
    def n(self):
        return self._n

    def last(self):
        return self.data.loc[-1]

    def n_last_out(self):
        return self._n_last_out

    # Make the dataframe accesible directly
    def __getitem__(self, *args):
        return self.data[:self.n()].__getitem__(*args)

    # Make the dataframe printable (but only the filled ones!)
    def __repr__(self):
        return self.data[:self.n()].__repr__()

    # Make the dataframe iterable over rows
    def __iter__(self):
        return self.data[:self.n()].iterrows()

    # Statistical computations
    def mean(self, first=None, last=None, derived=False):
        """
        Returns the (weighted) mean of the parameters in the chain,
        between `first` (default 0) and `last` (default last obtained),
        optionally including derived parameters if `derived=True` (default `False`).
        """
        return np.average(
            self[list(self.sampled_params)
                 +(list(self.derived_params) if derived else [])]
                [first:last].T,
            weights=self[_weight][first:last], axis=-1)

    def cov(self, first=None, last=None, derived=False):
        """
        Returns the (weighted) mean of the parameters in the chain,
        between `first` (default 0) and `last` (default last obtained),
        optionally including derived parameters if `derived=True` (default `False`).
        """
        return np.atleast_2d(np.cov(
            self[list(self.sampled_params)
                 +(list(self.derived_params) if derived else [])]
                [first:last].T,
            fweights=self[_weight][first:last]))

    # Saving and updating
    def get_driver(self, method):
        return getattr(self, method+separator+self.driver)

    # Create the file, write the header, etc.
    def out_prepare(self):
        self.get_driver("prepare")()
    # Dump/update collection
    def out_dump(self):
        self.get_driver("dump")()
    def out_update(self):
        self.get_driver("update")()

    # txt driver
    def prepare__txt(self):
        log.info("Sample collection to be written on '%s'", self.file_name)
        header = "#" + self.data[:0].to_csv(sep=" ", header=True)
        with open(self.file_name, "w") as out:
            out.write(header)
    def dump__txt(self):
        self.dump_slice__txt(0, self.n())
    def update__txt(self):
        self.dump_slice__txt(self.n_last_out(), self.n())
    def dump_slice__txt(self, n_min=None, n_max=None):
        if n_min == None or n_max == None:
            log.error("Needs to specify the limit n's to dump.")
            raise HandledException
        self._n_last_out = n_max
        self.data[n_min:n_max].to_csv(self.file_name, sep=" ", na_rep="nan", header=False,
                                      float_format="%.8g", index=False, mode="a")

    # dummy driver
    def prepare__dummy(self):
        pass
    def dump__dummy(self):
        pass
    def update__dummy(self):
        pass

    # Interface with getdist
    def as_getdist_mcsamples(self, derived=True, prior_and_lik=True,
                             first=None, last=None):
        # get names and labels (n.b.: getdist forcefully adds its own $'s)
        # sampled
        names = self.sampled_params + [p+"*" for p in self.derived_params]
        labels = self.labels.values()
        ranges = []
#        ranges = dict([(p,ls) for p,ls in zip(self.prior.names(),self.prior.limits())])
#         ranges.update(dict([(p,(self.likelihood.updated_info_params()[p].get("min"),
#                                    self.likelihood.updated_info_params()[p].get("max")))
#                                for p in self.likelihood.derived]))
        # prior and likelihood
        if prior_and_lik:
            names += [_minuslogprior+"*"] + [_chi2+"*"] + list(self.chi2_names)
            labels += ([r"-\log\pi", r"\chi^2"] +
                       [r"\chi^2_\mathrm{"+name+r"}" for name in self.chi2_names])
        # gather samples and create the getdist sample
        columns = [(("" if n.rstrip("*") not in self.derived_params else _derived_pre)+
                    n.rstrip("*")) for n in names]
        samples = self.data[:self.n()].as_matrix(columns=columns)
        return MCSamples(samples=samples[first:last],
                         weights=self.data[:self.n()][_weight].values[first:last],
                         loglikes=self.data[:self.n()][_minuslogpost].values[first:last],
                         names=names, labels=labels, ranges=ranges)
        
class OnePoint(Collection):
    """Wrapper of Collection to hold a single point, e.g. the current point of an MCMC."""
    def __init__(self, *args, **kwargs):
        kwargs["initial_size"] = 1
        Collection.__init__(self, *args, **kwargs)

    def __getitem__(self, columns):
        if isinstance(columns, basestring):
            columns = [columns]
        return self.data.as_matrix(columns=columns)[0]

    # Resets the counter, so the dataframe never fills up!
    def add(self, *args, **kwargs):
        Collection.add(self, *args, **kwargs)
        self._n = 0

    def increase_weight(self, increase):
        self.data["weight"] += increase
        
    def add_to_collection(self, collection):
        """Adds this point at the end of a given collection."""
        collection.add(
            self[self.sampled_params],
            derived=(self[[_derived_pre+p for p in self.derived_params]]
                     if self.derived_params else None),
            logpost=-self[_minuslogpost][0], weight=self[_weight][0],
            logprior=-self[_minuslogprior][0],
            logliks=-0.5*np.array(self[self.chi2_names]))
    
    # Make the dataframe printable (but only the filled ones!)
    def __repr__(self):
        return self.data[:1].__repr__()



"""
.. module:: one

:Synopsis: Mock constant value (1) likelihood
:Author: Jesus Torrado

Likelihoods that evaluates to 1. Useful to explore priors.

Options and defaults
--------------------

Simply copy this block in your input ``yaml`` file and modify whatever options you want
(you can delete the rest).

.. literalinclude:: ../cobaya/likelihoods/one/defaults.yaml
   :language: yaml

"""

# Python 2/3 compatibility
from __future__ import division

# Global
import numpy as np

# Local
from cobaya.likelihood import Likelihood

class one(Likelihood):
    """
    Likelihood that evaluates to 1.
    """

    def logp(self, **params_values):
        self.wait()
        return 0.



"""
.. module:: samplers.mcmc

:Synopsis: Blocked fast-slow Metropolis sampler (Lewis 1304.4473)
:Author: Antony Lewis (for the CosmoMC sampler), Jesus Torrado (for parts of the wrapper only)

.. |br| raw:: html

   <br />

.. note::
   **If you use this sampler, please cite it as:**
   |br|
   `A. Lewis and S. Bridle, "Cosmological parameters from CMB and other data: A Monte Carlo approach"
   (arXiv:astro-ph/0205436) <https://arxiv.org/abs/astro-ph/0205436>`_
   |br|
   `A. Lewis, "Efficient sampling of fast and slow cosmological parameters"
   (arXiv:1304.4473) <https://arxiv.org/abs/1304.4473>`_
   |br|
   If you use *fast-dragging*, you should also cite
   |br|
   `R.M. Neal, "Taking Bigger Metropolis Steps by Dragging Fast Variables"
   (arXiv:math/0502099) <https://arxiv.org/abs/math/0502099>`_


This is the Markov Chain Monte Carlo Metropolis sampler used by CosmoMC, and described in 
`Lewis, "Efficient sampling of fast and slow cosmological parameters" (arXiv:1304.4473) 
<https://arxiv.org/abs/1304.4473>`_.

The proposal pdf is a gaussian mixed with an exponential pdf in random directions, which is
more robust to misestimation of the width of the proposal than a pure gaussian. The scale width
of the proposal can be specified per parameter with the property ``proposal`` (it defaults to the
standard deviation of the reference pdf, if defined, or the prior's one, if not). However,
initial performance will be much better if you provide a covariance matrix, which overrides
the default proposal scale width set for each parameter.

A callback function can be specified through the ``callback_function`` option. In it, the
sampler instance is accessible as ``sampler_instance``, which has ``prior``, ``likelihood``
and (sample) ``collection`` as attributes.

Initial point and covariance of the proposal pdf
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The initial points for the chains are sampled from the *reference* pdf
(see :doc:`params_prior`). The reference pdf can be a fixed point, and in that case the
chain starts always from that same point. If there is no reference pdf defined for a
parameter, the initial sample is drawn from the prior instead.

Example *parameters* block:

.. code-block:: yaml
   :emphasize-lines: 10,17

   params:
     a:
      ref:
        min: -1
        max:  1
      prior:
        min: -2
        max:  2
      latex: \\alpha
      proposal: 0.5
     b:
      ref: 2
      prior:
        min: -1
        max:  4
      latex: \\beta
      proposal: 0.25
     c:
      ref:
        dist: norm
        loc: 0
        scale: 0.2        
      prior:
        min: -1
        max:  1
      latex: \\gamma

+ ``a`` -- the initial point of the chain is drawn from an uniform pdf between -1 and 1,
  and its proposal width is 0.5.
+ ``b`` -- the initial point of the chain is always 2,
  and its proposal width is 0.25.
+ ``c`` -- the initial point of the chain is drawn from a gaussian centred at 0
  with standard deviation 0.2; its proposal width is not specified, so it is taken to be 
  that of the reference pdf, 0.2.

A good initial covariance matrix for the proposal is critical for convergence.
It can be specified either with the property ``proposal`` of each parameter, as shown
above, or thorugh ``mcmc``'s property ``covmat``, as a file name (including path,
if not located at the invocation folder).
The first line of the ``covamt`` file must start with ``#``, followed by a list of parameter
names, separated by a space. The rest of the file must contain the covariance matrix,
one row per line. It does not need to contain the same parameters as the sampled ones: 
it overrides the ``proposal``'s (and adds covariances) for the sampled parameters,
and ignores the non-sampled ones.

An example for the case above::

   # a     b
     0.1   0.01
     0.01  0.2

In this case, internally, the final covariance matrix of the proposal would be::

   # a     b     c
     0.1   0.01  0
     0.01  0.2   0
     0     0     0.04

If the option `learn_proposal` is set to ``True``, the covariance matrix will be updated 
once in a while to accelerate convergence
(nb. convergence testing is only implemented for paralel chains right now).

If you are not sure that your posterior has one single mode, or if its shape is very
irregular, you should probably set ``learn_proposal: False``.

If you don't know how good your initial guess for starting point and covariance of the 
proposal are, it is a good idea to allow for a number of initial *burn in* samples,
e.g. 10 per dimension. This can be specified with the parameter ``burn_in``.
These samples will be ignored for all purposes (output, covergence, proposal learning...)


.. _mcmc_speed_hierarchy:

Taking advantage of a speed hierarchy
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The proposal pdf is *blocked* by speeds, i.e. it allows for efficient sampling of a mixture
of *fast* and *slow* parameters, such that we can avoid recomputing the parts of the
likelihood determined by slow parameters when sampling along the fast directions only.

Two different sampling schemes are available to take additional advantage from a speed
hierarchy:

- **Dragging the fast parameters:** implies a number of intermediate steps when
  jumping between fast+slow combinations, such that the jump in the fast parameters is
  optimised with respect to the jump in the slow parameters to explore any possible
  degeneracy betwen them.

- **Oversampling the fast parameters:** consists simply of taking a larger proportion of
  steps in the fast directions, since exploring their conditional distributions is cheap.

In general, the *dragging* method is the recommended one.
For a thorough description of both methods, see
`A. Lewis, "Efficient sampling of fast and slow cosmological parameters" (arXiv:1304.4473) <https://arxiv.org/abs/1304.4473>`_.

The relative speeds can be specified per likelihood/theory, with the option ``speed``.
It is given as factors of the speed of the slowest part, which is not specified.
If two or more likelihoods with different speeds share a parameter, said parameter is
assigned the slowest of their speeds.

For example:

.. code-block:: yaml

   theory:
     theory_code:
       speed: 2

   likelihood:
     lik_a:
     lik_b:
       speed: 4

Here, evaluating the likelihood ``lik_a`` is the slowest step, while the theory code and
the likelihood ``lik_b`` are both faster, with ``lik_b`` being faster than the theory code
(the absolute values of the speeds are ignored, only the relative ranking matters:
giving speeds 1,2,3 would have the same effect).

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
import six

# Global
from copy import deepcopy
from itertools import chain
import numpy as np

# Local
from cobaya.sampler import Sampler
from cobaya.mpi import get_mpi, get_mpi_size, get_mpi_rank, get_mpi_comm
from cobaya.collection import Collection, OnePoint
from cobaya.conventions import _weight, _p_proposal
from cobaya.samplers.mcmc.proposal import BlockedProposer
from cobaya.log import HandledException
from cobaya.tools import get_external_function

# Logger
import logging
log = logging.getLogger(__name__)


class mcmc(Sampler):

    def initialise(self):
        """Initialises the sampler:
        creates the proposal distribution and draws the initial sample."""
        log.info("Initialising")
        # Burning-in countdown -- the +1 accounts for the initial point (always accepted)
        self.burn_in_left = self.burn_in + 1
        # One collection per MPI process: `name` is the MPI rank + 1
        name = str(1 + (lambda r: r if r is not None else 0)(get_mpi_rank()))
        self.collection = Collection(
            self.parametrisation, self.likelihood, self.output, name=name)
        self.current_point = OnePoint(
            self.parametrisation, self.likelihood, self.output, name=name)
        # Use the standard steps by default
        self.get_new_sample = self.get_new_sample_metropolis
        # Create proposer -- speeds, fast-dragging/oversampling and initial covmat
        speeds, blocks = zip(*self.likelihood.speeds_of_params().items())
        # Turn parameter names into indices
        blocks = [[self.parametrisation.sampled_params().keys().index(p) for p in b]
                  for b in blocks]
        if self.oversample and (self.drag_nfast_times or self.drag_interp_steps):
            log.error("Choose either oversampling or fast-dragging, not both.")
            raise HandledException
        if self.oversample:
            self.oversampling_factors = [int(np.round(s/speeds[0])) for s in speeds]
            if len(set(self.oversampling_factors)) == 1:
                log.error("All likelihood speeds are similar: no oversampling possible.")
                raise HandledException
            self.effective_max_samples = (
                sum([len(b)*f for b,f in zip(blocks,self.oversampling_factors)]) /
                len(self.parametrisation.sampled_params()))
            self.n_slow = len(blocks[0])
        elif self.drag_interp_steps or self.drag_nfast_times:
            if len(set(speeds)) == 1:
                log.error("All likelihoods speeds are equal: no fast_dragging possible.")
                raise HandledException
            if self.drag_nfast_times and self.drag_interp_steps:
                log.error("To specify the number of dragging interpolating steps, use "
                          "*either* `drag_nfast_times` or `drag_interp_steps`, not both.")
                raise HandledException
            if self.max_speed_slow < min(speeds) or self.max_speed_slow >= max(speeds):
                log.error("The maximum speed considered slow, `max_speed_slow`, must be "
                          "%g <= `max_speed_slow < %g, and is %g",
                          min(speeds), max(speeds), self.max_speed_slow)
                raise HandledException
            self.i_last_slow_block = (i for i,speed in enumerate(list(speeds))
                                      if speed > self.max_speed_slow).next() - 1
            fast_params = [self.parametrisation.sampled_params().keys()[i]
                           for i in chain(*blocks[1+self.i_last_slow_block:])]
            self.effective_max_samples = self.max_samples
            self.n_slow = sum(len(blocks[i]) for i in range(1+self.i_last_slow_block))
            if self.drag_nfast_times:
                self.drag_interp_steps = max(
                    2, int(self.drag_nfast_times*len(fast_params))+1)
            self.get_new_sample = self.get_new_sample_dragging
            log.info("Using fast dragging over %d slow parameters, "
                     "with %d interpolating steps on fast parameters %r",
                     self.n_slow, self.drag_interp_steps, fast_params)
        else:
            self.effective_max_samples = self.max_samples
            self.n_slow = len(self.parametrisation.sampled_params())
        self.proposer = BlockedProposer(
            blocks, oversampling_factors=getattr(self, "oversampling_factors", None),
            i_last_slow_block=getattr(self, "i_last_slow_block", None),
            propose_scale=self.propose_scale)
        # Build the initial covariance matrix of the proposal
        covmat = self.initial_proposal_covmat()
        log.info("Sampling with covariance matrix:")
        log.info("%r", covmat)
        self.proposer.set_covariance(covmat)
        # Prepare callback function
        if self.callback_function is not None:
            self.callback_function_callable = (
                get_external_function(self.callback_function))

    def initial_proposal_covmat(self):
        """
        Build the initial covariance matrix, using the data provided, in descending order
        of priority:
        1. "covmat" field in the "mcmc" sampler block.
        2. "proposal" field for each parameter.
        3. variance of the reference pdf.
        4. variance of the prior pdf.

        The covariances between parameters when both are present in a covariance matrix
        provided through option 1 are preserved. All other covariances are assumed 0.
        """
        params, params_infos = zip(*self.parametrisation.sampled_params().items())
        covmat = np.diag([np.nan]*len(params))
        # If given, load and test the covariance matrix
        if isinstance(self.covmat, six.string_types):
            try:
                with open(self.covmat, "r") as file_covmat:
                    header = file_covmat.readline()
                loaded_covmat = np.loadtxt(self.covmat)
            except TypeError:
                log.error("The property 'covmat' must be a file name,"
                          "but it's '%s'.", str(self.covmat))
                raise HandledException
            except IOError:
                log.error("Can't open covmat file '%s'.", self.covmat)
                raise HandledException
            if header[0] != "#":
                log.error("The first line of the covmat file '%s' "
                          "must be one list of parameter names separated by spaces "
                          "and staring with '#', and the rest must be a square matrix, "
                          "with one row per line.", self.covmat)
                raise HandledException
            loaded_params = header.strip("#").strip().split()
        elif hasattr(self.covmat, "__getitem__"):
            if not self.covmat_params:
                log.error("If a covariance matrix is passed as a numpy array, "
                          "you also need to pass the parameters it corresponds to "
                          "via 'covmat_params: [name1, name2, ...]'.")
                raise HandledException
            loaded_params = self.covmat_params
            loaded_covmat = self.covmat
        if self.covmat is not None:
            if len(loaded_params) != len(set(loaded_params)):
                log.error("There are duplicated parameters in the header of the "
                          "covmat file '%s' ", self.covmat)
                raise HandledException
            if len(loaded_params) != loaded_covmat.shape[0]:
                log.error("The number of parameters in the header of '%s' and the "
                          "dimensions of the matrix do not coincide.", self.covmat)
                raise HandledException
            if not (np.allclose(loaded_covmat.T, loaded_covmat) and
                    np.all(np.linalg.eigvals(loaded_covmat) > 0)):
                log.error("The covmat loaded from '%s' is not a positive-definite, "
                          "symmetric square matrix.", self.covmat)
                raise HandledException
            # Fill with parameters in the loaded covmat
            loaded_params_used = set(loaded_params).intersection(set(params))
            if not loaded_params_used:
                log.error("A proposal covariance matrix has been loaded, but none of its "
                          "parameters are actually sampled here. Maybe a mismatch between"
                          " parameter names in the covariance matrix and the input file?")
                raise HandledException
            indices_used, indices_sampler = np.array(
                [[loaded_params.index(p),params.index(p)]
                 for p in loaded_params if p in loaded_params_used]).T
            covmat[np.ix_(indices_sampler,indices_sampler)] = (
                loaded_covmat[np.ix_(indices_used,indices_used)])
            log.info("Covariance matrix loaded for params %r", list(loaded_params_used))
        # Fill gaps with "proposal" property, if present, otherwise ref (or prior)
        where_nan = np.isnan(covmat.diagonal())
        if np.any(where_nan):
            covmat[where_nan, where_nan] = np.array(
                [info.get(_p_proposal, np.nan)**2 for info in params_infos])[where_nan]
            # we want to start learning the covmat earlier
            log.info("Covariance matrix " +
                     ("not present" if np.all(where_nan) else "not complete") + ". "
                     "We will start learning the covariance of the proposal earlier: "
                     "R-1 = %g (was %g).", self.learn_proposal_Rminus1_max_early,
                     self.learn_proposal_Rminus1_max)
            self.learn_proposal_Rminus1_max = self.learn_proposal_Rminus1_max_early
        where_nan = np.isnan(covmat.diagonal())
        if np.any(where_nan):
            covmat[where_nan, where_nan] = (
                self.prior.reference_covmat().diagonal()[where_nan])
        assert not np.any(np.isnan(covmat))
        return covmat

    def run(self):
        """
        Runs the sampler.
        """
        # Get first point, to be discarded -- not possible to determine its weight
        # Still, we need to compute derived parameters, since, as the proposal "blocked",
        # we may be saving the initial state of some block.
        initial_point = self.prior.reference(max_tries=self.max_tries)
        logpost, _, _, derived = self.logposterior(initial_point)
        self.current_point.add(initial_point, derived=derived, logpost=logpost)
        log.info("Initial point:\n %r ",self.current_point)
        # Main loop!
        self.converged = False
        log.info("Sampling!")
        while self.n() < (self.effective_max_samples) and not(self.converged):
            self.get_new_sample()
            # Callback function
            if (hasattr(self, "callback_function_callable") and
                    not(max(self.n(),1)%self.callback_every) and
                    self.current_point[_weight] == 1):
                self.callback_function_callable(self)
            # Checking convergence and (optionally) learning the covmat of the proposal
            if self.check_all_ready():
                self.check_convergence_and_learn_proposal()
        # Make sure the last batch of samples ( < output_every ) are written
        self.collection.out_update()
        if not get_mpi_rank():
            log.info("Sampling complete after %d accepted steps.", self.n())

    def n(self, burn_in=False):
        """
        Returns the total number of steps taken, including or not burn-in steps depending
        on the value of the `burn_in` keyword.
        """
        return self.collection.n() + (
            0 if not burn_in else self.burn_in - self.burn_in_left + 1)

    def get_new_sample_metropolis(self):
        """
        Draws a new trial point from the proposal pdf and checks whether it is accepted:
        if it is accepted, it saves the old one into the collection and sets the new one
        as the current state; if it is rejected increases the weight of the current state
        by 1.

        Returns:
           ``True`` for an accepted step, ``False`` for a rejected one.
        """
        trial = deepcopy(self.current_point[self.parametrisation.sampled_params()])
        self.proposer.get_proposal(trial)
        logpost_trial, logprior_trial, logliks_trial, derived = self.logposterior(trial)
        accept = self.metropolis_accept(logpost_trial, -self.current_point["minuslogpost"])
        self.process_accept_or_reject(accept, trial, derived,
                                      logpost_trial, logprior_trial, logliks_trial)
        return accept

    def get_new_sample_dragging(self):
        """
        Draws a new trial point in the slow subspace, and gets the corresponding trial
        in the fast subspace by "dragging" the fast parameters.
        Finally, checks the acceptance of the total step using the "dragging" pdf:
        if it is accepted, it saves the old one into the collection and sets the new one
        as the current state; if it is rejected increases the weight of the current state
        by 1.

        Returns:
           ``True`` for an accepted step, ``False`` for a rejected one.
        """
        # Prepare starting and ending points *in the SLOW subspace*
        # "start_" and "end_" mean here the extremes in the SLOW subspace
        start_slow_point = self.current_point[self.parametrisation.sampled_params()]
        start_slow_logpost = -self.current_point["minuslogpost"]
        end_slow_point = deepcopy(start_slow_point)
        self.proposer.get_proposal_slow(end_slow_point)
        # Save derived paramters of delta_slow jump, in case I reject all the dragging
        # steps but accept the move in the slow direction only
        end_slow_logpost, end_slow_logprior, end_slow_logliks, derived = (
            self.logposterior(end_slow_point))
        if end_slow_logpost == -np.inf:
            self.current_point.increase_weight(1)
            return False
        # trackers of the dragging
        current_start_point = start_slow_point
        current_end_point   = end_slow_point
        current_start_logpost = start_slow_logpost
        current_end_logpost   = end_slow_logpost
        current_end_logprior  = end_slow_logprior
        current_end_logliks   = end_slow_logliks
        # accumulators for the "dragging" probabilities to be metropolist-tested
        # at the end of the interpolation
        start_drag_logpost_acc = start_slow_logpost
        end_drag_logpost_acc   = end_slow_logpost
        # start dragging
        for i_step in range(1, 1+self.drag_interp_steps):
            log.debug("Dragging step: %d", i_step)
            # take a step in the fast direction in both slow extremes
            delta_fast = np.zeros(len(current_start_point))
            self.proposer.get_proposal_fast(delta_fast)
            proposal_start_point = deepcopy(current_start_point)
            proposal_start_point += delta_fast
            proposal_end_point    = deepcopy(current_end_point)
            proposal_end_point   += delta_fast
            # get the new extremes for the interpolated probability
            # (reject if any of them = -inf; avoid evaluating both if just one fails)
            # Force the computation of the (slow blocks) derived params at the starting
            # point, but discard them, since they contain the starting point's fast ones,
            # not used later -- save the end point's ones.
            proposal_start_logpost = self.logposterior(proposal_start_point)[0]
            proposal_end_logpost, proposal_end_logprior, \
                proposal_end_logliks, derived_proposal_end = (
                    self.logposterior(proposal_end_point)
                    if proposal_start_logpost > -np.inf
                    else (-np.inf, None, [], []))
            if proposal_start_logpost > -np.inf and proposal_end_logpost > -np.inf:
                # create the interpolated probability and do a Metropolis test
                frac = i_step / self.drag_interp_steps
                proposal_interp_logpost = ((1-frac)*proposal_start_logpost
                                             +frac *proposal_end_logpost)
                current_interp_logpost  = ((1-frac)*current_start_logpost
                                             +frac *current_end_logpost)
                accept_drag = self.metropolis_accept(proposal_interp_logpost,
                                                     current_interp_logpost)
            else:
                accept_drag = False
            log.debug("Dragging step: %s", ("accepted" if accept_drag else "rejected"))
            # If the dragging step was accepted, do the drag
            if accept_drag:
                current_start_point   = proposal_start_point
                current_start_logpost = proposal_start_logpost
                current_end_point     = proposal_end_point
                current_end_logpost   = proposal_end_logpost
                current_end_logprior  = proposal_end_logprior
                current_end_logliks   = proposal_end_logliks
                derived = derived_proposal_end
            # In any case, update the dragging probability for the final metropolis test
            start_drag_logpost_acc += current_start_logpost
            end_drag_logpost_acc   += current_end_logpost
        # Test for the TOTAL step
        accept = self.metropolis_accept(end_drag_logpost_acc/self.drag_interp_steps,
                                        start_drag_logpost_acc/self.drag_interp_steps)
        self.process_accept_or_reject(
            accept, current_end_point, derived,
            current_end_logpost, current_end_logprior, current_end_logliks)
        log.debug("TOTAL step: %s", ("accepted" if accept else "rejected"))
        return accept
    
    def metropolis_accept(self, logp_trial, logp_current):
        """
        Symmetric-proposal Metropolis-Hastings test.

        Returns:
           ``True`` or ``False``.
        """
        if logp_trial == -np.inf:
            return False
        elif logp_trial > logp_current:
            return True
        else:
            return np.random.exponential() > (logp_current - logp_trial)

    def process_accept_or_reject(self, accept_state, trial=None, derived=None,
                                 logpost_trial=None, logprior_trial=None, logliks_trial=None):
        """Processes the acceptance/rejection of the new point."""
        if accept_state:
            # add the old point to the collection (if not burning or initial point)
            if self.burn_in_left <= 0:
                self.current_point.add_to_collection(self.collection)
                log.debug("New sample, #%d: \n   %r", self.n(), self.current_point)
                if self.n()%self.output_every == 0:
                    self.collection.out_update()
            else:
                self.burn_in_left -= 1
                log.debug("Burn-in sample:\n   %r", self.current_point)
                if self.burn_in_left == 0:
                    log.info("Finished burn-in phase: discarded %d accepted steps.",
                             self.burn_in)
            # set the new point as the current one, with weight one
            self.current_point.add(trial, derived=derived, weight=1, logpost=logpost_trial,
                                   logprior=logprior_trial, logliks=logliks_trial)
        else: # not accepted
            self.current_point.increase_weight(1)
            # Failure criterion: chain stuck!
            if self.current_point[_weight] > self.max_tries:
                self.collection.out_update()
                log.error(
                    "The chain has been stuck for %d attempts. "
                    "Stopping sampling. If this has happened often, try improving your"
                    " reference point/distribution.", self.max_tries)
                raise HandledException

    # Functions to check convergence and learn the covariance of the proposal distribution

    def check_all_ready(self):
        """
        Checks if the chain(s) is(/are) ready to check convergence and, if requested,
        learn a new covariance matrix for the proposal distribution.
        """
        msg_ready = (("Ready to" if get_mpi() or self.learn_proposal else "") +
                     (" check convergence" if get_mpi() else "") +
                     (" and" if get_mpi() and self.learn_proposal else "") +
                     (" learn a new proposal covmat" if self.learn_proposal else ""))
        # If *just* (weight==1) got ready to check+learn
        if (self.n() > 0 and self.current_point[_weight]==1 and
            not (self.n()%(self.check_every_dimension_times*self.n_slow))):
            log.info("Checkpoint: %d samples accepted.", self.n())
            # If not MPI, we are ready
            if not get_mpi():
                if msg_ready:
                    log.info(msg_ready)
                return True
            # If MPI, tell the rest that we are ready -- we use a "gather"
            # ("reduce" was problematic), but we are in practice just pinging
            if not hasattr(self, "req"): # just once!
                self.all_ready = np.empty(get_mpi_size())
                self.req = get_mpi_comm().Iallgather(
                    np.array([1.]), self.all_ready)
                log.info(msg_ready + " (waiting for the rest...)")
        # If all processes are ready to learn (= communication finished)
        if (self.req.Test() if hasattr(self, "req") else False):
            # Sanity check: actually all processes have finished
            assert np.all(self.all_ready == 1), (
                "This should not happen! Notify the developers. (Got %r)", self.all_ready)
            if get_mpi_rank() == 0:
                log.info("All chains are r"+msg_ready[1:])
            delattr(self, "req")
            # Just in case, a barrier here
            get_mpi_comm().barrier()
            return True
        return False

    def check_convergence_and_learn_proposal(self):
        """
        Checks the convergence of the sampling process (MPI only), and, if requested,
        learns a new covariance matrix for the proposal distribution from the covariance
        of the last samples.
        """
        # Compute and gather means, covs and cl limits of last half of chains
        mean = self.collection.mean(first=int(self.n()/2))
        cov  = self.collection.cov (first=int(self.n()/2))
        # Turn off logging of warnings momentarily, so getdist won't complain innecessarily
        logging.disable(logging.WARNING)
        mcsamples = self.collection.as_getdist_mcsamples(
            derived=False, prior_and_lik=False, first=int(self.n()/2))
        logging.disable(logging.NOTSET)
        limit = np.array([[mcsamples.confidence(i, limfrac=self.Rminus1_cl_level/2., upper=lim)
                            for i in range(self.prior.d())] for lim in [False, True]]).T
        Ns, means, covs, limits = map(
            lambda x: np.array((get_mpi_comm().gather(x) if get_mpi() else [x])),
            [self.n(), mean, cov, limit])
        # Compute convergence diagnostics
        if get_mpi():
            if get_mpi_rank() == 0:
                # "Within" or "W" term -- our "units" for assessing convergence
                # and our prospective new covariance matrix
                mean_of_covs = np.average(covs, weights=Ns, axis=0)
                # "Between" or "B" term
                # We don't weight with the number of samples in the chains here:
                # shorter chains will likely be outliers, and we want to notice them
                cov_of_means = np.cov(means.T) #, fweights=Ns)
                # For numerical stability, we turn mean_of_covs into correlation matrix:
                #   rho = (diag(Sigma))^(-1/2) * Sigma * (diag(Sigma))^(-1/2)
                # and apply the same transformation to the mean of covs (same eigenvals!)
                diagSinvsqrt = np.diag(np.power(np.diag(cov_of_means), -0.5))
                corr_of_means     = diagSinvsqrt.dot(cov_of_means).dot(diagSinvsqrt)
                norm_mean_of_covs = diagSinvsqrt.dot(mean_of_covs).dot(diagSinvsqrt)
                # Cholesky of (normalised) mean of covs and eigvals of Linv*cov_of_means*L
                L = np.linalg.cholesky(norm_mean_of_covs)
                Linv = np.linalg.inv(L)
                eigvals = np.linalg.eigvalsh(Linv.dot(corr_of_means).dot(Linv.T))                
                Rminus1 = max(np.abs(eigvals))
                # For real square matrices, a possible def of the cond number is:
                condition_number = Rminus1/min(np.abs(eigvals))
                log.debug("Condition number = %g", condition_number)
                log.debug("Eigenvalues = %r", eigvals)
                log.info("Convergence of means: R-1 = %f after %d samples", Rminus1, self.n())
#                assert np.all(eigvals >= 0), (
#                    "Negative eigenvectors. This should not happen. "
#                    "Check the condition number of the diagonalised matrix.")
                # Have we converged in means? (criterion must be fulfilled twice in a row)
                if (max(Rminus1, getattr(self, "Rminus1_last", np.inf)) < self.Rminus1_stop):
                    # Check the convergence of the limits of the confidence intervals
                    # Same as R-1, but with the rms deviation from the mean limit
                    # in units of the mean standard deviation of the chains
                    Rminus1_cl = np.std(limits, axis=0).T/np.sqrt(np.diag(mean_of_covs))
                    log.debug("Normalised std's of limits = %r", Rminus1_cl)
                    log.info("Convergence of limits: R-1 = %f after %d samples", 
                             np.max(Rminus1_cl), self.n())
                    if np.max(Rminus1_cl) < self.Rminus1_cl_stop:
                        self.converged=True
                        log.info("The run has converged!")
            # Broadcast and save the convergence status and the last R-1 of means
            self.Rminus1_last = get_mpi_comm().bcast(
                (Rminus1 if not get_mpi_rank() else None), root=0)
            self.converged = get_mpi_comm().bcast(
                (self.converged if not get_mpi_rank() else None), root=0)
        else: # No MPI
            pass
        # Do we want to learn a better proposal pdf?
        if self.learn_proposal and not self.converged:
            # update iff (not MPI, or MPI and "good" Rminus1)
            if get_mpi():
                good_Rminus1 = (self.Rminus1_last < self.learn_proposal_Rminus1_max and
                                self.Rminus1_last > self.learn_proposal_Rminus1_min)
                if not good_Rminus1:
                    if not get_mpi_rank():
                        log.info("Bad convergence statistics: waiting until the next checkpoint.")
                    return
            if get_mpi():
                if get_mpi_rank():
                    mean_of_covs = np.empty((self.prior.d(),self.prior.d()))
                get_mpi_comm().Bcast(mean_of_covs, root=0)
            elif not get_mpi():
                mean_of_covs = covs[0]
            self.proposer.set_covariance(mean_of_covs)
            if not get_mpi_rank():
                log.info("Updated covariance matrix of proposal pdf.")
                log.debug("%r", mean_of_covs)

    # Finally: returning the computed products ############################################

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.

        Returns:
           The sample ``Collection`` containing the accepted steps.
        """
        return {"sample": self.collection}



"""
.. module:: install

:Synopsis: Tools and script to install the modules requested in the given input.
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
import sys
import logging
from cobaya.log import logger_setup, HandledException
from importlib import import_module

# Local
from cobaya.tools import get_folder, make_header
from cobaya.input import get_modules
from cobaya.conventions import package

def install(*infos, **kwargs):
    log = logging.getLogger(__name__)
    path = kwargs.get("path", ".")
    abspath = os.path.abspath(path)
    if not os.path.exists(abspath):
        log.error("The given path, %s, must exist, but it doesn't."%abspath)
        raise HandledException
    force = kwargs.get("force", False)
    for kind, modules in get_modules(*infos).iteritems():
        # Create folder for kind
        kindpath = os.path.join(abspath, get_folder("", kind, sep="/", absolute=False))
        if not os.path.exists(kindpath):
            os.makedirs(kindpath)
        for module in modules:
            installpath = os.path.join(kindpath, module)
            print make_header(kind, module)
            module_folder = get_folder(module, kind, sep=".", absolute=False)
            imported_module = import_module(module_folder, package=package)
            is_installed = getattr(imported_module, "is_installed", None)
            if is_installed is None:
                print "Not and external module: nothing to do.\n"
                continue
            if is_installed(path=installpath):
                print "External module appears to be installed."
                if force:
                    print "Forcing re-installation, as requested."
                else:
                    print "Doing nothing.\n"
                    continue
            success = imported_module.install(path=installpath, force=force)
            if success:
                print "Successfully installed!\n"
            else:
                log.error("Installation failed! Look at the error messages above. "
                          "Solve them and try again, or, if you are unable to solve, "
                          "install this module manually.")
                continue
            # test installation
            if not is_installed(path=installpath):
                log.error("Installation apparently worked, but the subsequent installation"
                          " test failed! Look at the error messages above. "
                          "Solve them and try again, or, if you are unable to solve, "
                          "install this module manually.")

# Add --user flag to pip, if needed: when not in TRAVIS, Anaconda or a virtual env
def user_flag_if_needed():
    if ("TRAVIS" not in os.environ  # Travis
        and not any([(s in sys.version) for s in ["conda", "Continuum"]])  # Anaconda
        and not hasattr(sys, 'real_prefix')  # Virtual environment (virtualenv)
        and getattr(sys, 'base_prefix', sys.prefix) == sys.prefix):  # Idem (pyvenv)
        return ["--user"]
    return []

# Command-line script
def install_script():
    from cobaya.mpi import get_mpi_rank
    if not get_mpi_rank():
        # Configure the logger ASAP
        logger_setup()
        log = logging.getLogger(__name__)
        # Parse arguments
        import argparse
        parser = argparse.ArgumentParser(
            description="Cobaya's installation tool for external modules.")
        parser.add_argument("files", action="store", nargs="+", metavar="input_file.yaml",
                            help="One or more input files.")
        parser.add_argument("-p", "--path",
                            action="store", nargs=1, default=".", metavar=("/some/path"),
                            help="Desired path where to install external modules.")
        parser.add_argument("-f", "--force", action="store_true", default=False,
                            help="Force re-installation of apparently installed modules.")
        arguments = parser.parse_args()
        from cobaya.input import load_input
        infos = [load_input(f) for f in arguments.files]
        # Launch installer
        install(*infos, path=arguments.path[0], force=arguments.force)



"""
.. module:: run

:Synopsis: Main scripts to run the sampling
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
import six

# Global
from copy import deepcopy


def run(info):

    assert hasattr(info, "items"), (
        "The agument of `run` must be a dictionary with the info needed for the run. "
        "If you were trying to pass an input file instead, load it first with "
        "`cobaya.input.load_input`.")

    # Import names
    from cobaya.conventions import _likelihood, _prior, _params
    from cobaya.conventions import _theory, _sampler, _path_install
    from cobaya.conventions import _debug, _debug_file, _output_prefix

    # Configure the logger ASAP
    from cobaya.log import logger_setup
    logger_setup(info.get(_debug), info.get(_debug_file))
    
    # Debug (lazy call)
    import logging
    if logging.root.getEffectiveLevel() <= logging.DEBUG:
        from cobaya.yaml_custom import yaml_dump
        logging.getLogger(__name__).debug(
            "Input info (dumped to YAML):\n%s", yaml_dump(info))
    
    # Import general classes
    from cobaya.prior import Prior
    from cobaya.sampler import get_Sampler as Sampler

    # Import the functions and classes that need MPI wrapping
    from cobaya.mpi import import_MPI
#    Likelihood = import_MPI(".likelihood", "LikelihoodCollection")
    from cobaya.likelihood import LikelihoodCollection as Likelihood

    # Initialise output, if requiered
    do_output = info.get(_output_prefix)
    if do_output:
        Output = import_MPI(".output", "Output")
        output = Output(info)
    else:
        from cobaya.output import Output_dummy
        output = Output_dummy(info)

    # Create the full input information, including defaults for each module
    from cobaya.input import get_full_info
    full_info = get_full_info(info)
    output.set_full_info(full_info)
    if logging.root.getEffectiveLevel() <= logging.DEBUG:
        logging.getLogger(__name__).debug(
            "Updated info (dumped to YAML):\n%s", yaml_dump(full_info))

    # Set the path of the installed modules, if given
    from cobaya.tools import set_path_to_installation
    set_path_to_installation(info.get(_path_install))

    # Initialise parametrisation, likelihoods and prior
    from cobaya.parametrisation import Parametrisation
    with Parametrisation(full_info[_params]) as par:
        with Prior(par, full_info.get(_prior)) as prior:
            with Likelihood(full_info[_likelihood], par, full_info.get(_theory)) as lik:
                with Sampler(full_info[_sampler], par, prior, lik, output) as sampler:
                    # Save the full, defaults-populated info
                    output.dump_info()
                    # Sample!
                    sampler.run()
    # For scripted calls
    return deepcopy(full_info), sampler.products()


# Command-line script
def run_script():
    from cobaya.mpi import import_MPI
    load_input = import_MPI(".input", "load_input")
    import argparse
    parser = argparse.ArgumentParser(description="Cobaya's run script.")
    parser.add_argument("input_file", nargs=1, action="store", metavar="input_file.yaml",
                        help="An input file to run.")
    parser.add_argument("-p", "--path",
                        action="store", nargs="+", metavar=("/some/path"),
                        help="Path where modules were automatically installed.")
    args = parser.parse_args()
    info = load_input(args.input_file[0])
    # solve path
    from cobaya.conventions import _path_install
    path_cmd = (lambda x: x[0] if x else None)(getattr(args, "path"))
    path_input = info.get(_path_install)
    if path_cmd:
        if path_input:
            raise ValueError("CONFLICT: "
                      "You have specified a modules folder both in the command line "
                      "('%s') and the input file ('%s'). There should only be one."%(
                          path_cmd, path_input))
        info[_path_install] = path_cmd
    run(info)



"""
.. module:: external

:Synopsis: External theory code, that would implement the necessary methods for the likelihoods.
:Author: Jesus Torrado

.. |br| raw:: html

   <br />

This dummy module simply interfaces an external class, whose methods are called instead of
those of a real theory class.

Usage
-----

Just pass an instance of a class thorugh the parameter ``instance``,
 containing the necessary methods for the likelihood to retrieve observables.

.. todo::

   Add example!


"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import inspect

# Local
from cobaya.theory import Theory
from cobaya.tools import yaml_custom_load

class external(Theory):

    def initialise(self):
        # Make the methods of the external instance callable from this class
        for name, method in inspect.getmembers(self.instance, predicate=inspect.ismethod):
            if name == "__init__":
                continue
            setattr(self, name, method)



"""
.. module:: mpi

:Synopsis: Manages MPI parallelisation transparently
:Author: Jesus Torrado

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os

# Local
from cobaya.conventions import package

# Vars to keep track of MPI parameters
_mpi = -1
_mpi_size = -1
_mpi_comm = -1
_mpi_rank = -1

def get_mpi():
    """
    Import and returns the MPI object, or None of not running with MPI.

    Can be used as a boolean test if MPI is present.
    """
    global _mpi
    if _mpi == -1:
        if os.environ.get("OMPI_COMM_WORLD_SIZE"):
            from mpi4py import MPI
            _mpi = MPI
        else:
            _mpi = None
    return _mpi

def get_mpi_size():
    """
    Returns the number of MPI processes that have been invoked,
    or 0 if not running with MPI.
    """
    global _mpi_size
    if _mpi_size == -1:
        _mpi_size = int(os.environ.get("OMPI_COMM_WORLD_SIZE", 0))
    return _mpi_size

def get_mpi_comm():
    """
    Returns the MPI communicator, or `None` if not running with MPI.
    """
    global _mpi_comm
    if _mpi_comm == -1:
        _mpi_comm = getattr(get_mpi(), "COMM_WORLD", None)
    return _mpi_comm

def get_mpi_rank():
    """
    Returns the rank of the current MPI process:
        * None: not running with MPI
        * Z>=0: process rank, when running with MPI

    Can be used as a boolean that returns `False` for both the root process,
    if running with MPI, or always for a single process; thus, everything under
    `if not(get_mpi_rank()):` is run only *once*.
    """
    global _mpi_rank
    if _mpi_rank == -1:
        _mpi_rank = getattr(get_mpi_comm(), "Get_rank", lambda: None)()
    return _mpi_rank

def import_MPI(module, target):
    """Import helper for MPI wrappers."""
    from importlib import import_module
    target_name = target
    if get_mpi_rank() >= 0:
        target_name = target+"_MPI"
    return getattr(import_module(module, package=package), target_name)



"""
.. module:: samplers.mcmc.proposal

:Synopsis: proposal distributions
:Author: Antony Lewis (from CosmoMC)

Using the covariance matrix to give the proposal directions typically
significantly increases the acceptance rate and gives faster movement
around parameter space.

We generate a random basis in the eigenvectors, then cycle through them
proposing changes to each, then generate a new random basis.
The distance proposal in the random direction is given by a two-D Gaussian
radial function mixed with an exponential, which is quite robust to wrong width estimates

See http://cosmologist.info/notes/CosmoMC.pdf
"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import numpy as np
import scipy
import numpy
from itertools import chain

# Local
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)


class IndexCycler(object):
    def __init__(self, n):
        self.n = n
        self.loop_index = -1


class CyclicIndexRandomizer(IndexCycler):
    def next(self):
        """
        Get the next random index

        :return: index
        """
        self.loop_index = (self.loop_index + 1) % self.n
        if self.loop_index == 0:
            self.indices = np.random.permutation(self.n)
        return self.indices[self.loop_index]


class RandDirectionProposer(IndexCycler):
    def propose_vec(self, scale=1):
        """
        propose a random n-dimension vector

        :param scale: units for the distance
        :return: array with vector
        """
        self.loop_index = (self.loop_index + 1) % self.n
        if self.loop_index == 0:
            if self.n > 1:
                self.R = scipy.stats.special_ortho_group.rvs(self.n)
            else:
                self.R = np.eye(1)*numpy.random.choice((-1,1))
        return self.R[:, self.loop_index] * self.propose_r() * scale

    def propose_r(self):
        """
        Radial proposal. By default a mixture of an exponential and 2D Gaussian radial
        proposal (to make wider tails and more mass near zero, so more robust to scale
        misestimation)

        :return: random distance (unit scale)
        """
        if np.random.uniform() < 0.33:
            return np.random.exponential()
        else:
            return np.linalg.norm(np.random.normal(size=min(self.n, 2)))


class BlockedProposer(object):
    def __init__(self, parameter_blocks, oversampling_factors=None,
                 i_last_slow_block=None, propose_scale=2.4):
        """
        Proposal density for fast and slow parameters, where parameters are
        grouped into blocks which are changed at the same time.

        :param parameter_blocks: list of arrays of parameter indices in each block,
            with blocks sorted by ascending speed.
        :param oversampling_factors: list of oversampling factors for each block.
        :param i_last_slow_block: index of the last block considered slow.
            By default, all blocks are considered slow.
        :param propose_scale: overal scale for the proposal.
        """
        self.propose_scale = propose_scale
        if oversampling_factors is None:
            self.oversampling_factors = [1 for _ in parameter_blocks]
        else:
            if len(oversampling_factors) != len(parameter_blocks):
                log.error("List of oversampling factors has a different length that "
                          "list of blocks: %d vs %d.",
                          len(oversampling_factors), len(parameter_blocks))
                raise HandledException
            self.oversampling_factors = [max(1,int(f)) for f in oversampling_factors]
        if i_last_slow_block is None:
            i_last_slow_block = len(parameter_blocks)-1
        else:
            if i_last_slow_block > len(parameter_blocks)-1:
                log.error("The index given for the last slow block, %d, is not valid: "
                          "there are only %d blocks.",
                          i_last_slow_block, len(parameter_blocks))
                raise HandledException
        n_all = sum([len(b) for b in parameter_blocks])
        n_slow = sum([len(b) for b in parameter_blocks[:1+i_last_slow_block]])
        if set(list(chain(*parameter_blocks))) != set(range(n_all)):
            log.error("The blocks do not contain all the parameter indices.")
            raise HandledException
        # Mapping between internal indices, sampler parameter indices and blocks:
        # let i=0,1,... be the indices of the parameters for the sampler,
        # and j=0,1,... be the indices of the parameters *as given to the proposal*
        # i.e. if passed blocks=[[1,2],[0]] (those are the i's),
        # then the j's are [0 (for 1), 1 (for 2), 2 (for 0)].
        # iblock is the index of the blocks
        self.i_of_j = np.array(list(chain(*parameter_blocks)))
        self.iblock_of_j = list(
            chain(*[[iblock]*len(b) for iblock,b in enumerate(parameter_blocks)]))
        # Creating the blocked proposers
        self.proposer = [RandDirectionProposer(len(b)) for b in parameter_blocks]
        # Starting j index of each block
        self.j_start = [len(list(chain(*parameter_blocks[:iblock])))
                        for iblock,b in enumerate(parameter_blocks)]
        # Parameter cyclers, cycling over the j's
        self.cycler_all = CyclicIndexRandomizer(n_all)
        # These ones are used by fast dragging only
        self.cycler_slow = CyclicIndexRandomizer(n_slow)
        self.cycler_fast = CyclicIndexRandomizer(n_all - n_slow)
        # Samples left to draw from the current block
        self.samples_left = 0

    def d(self):
        return len(self.i_of_j)

    def get_proposal(self, P):
        # if a block has been chosen
        if self.samples_left:
            self.get_block_proposal(P, self.current_iblock)
            self.samples_left -= 1
        # otherwise, choose a block
        else:
            self.current_iblock = self.iblock_of_j[self.cycler_all.next()]
            self.samples_left = self.oversampling_factors[self.current_iblock]
            self.get_proposal(P)

    def get_proposal_slow(self, P):
        current_iblock_slow = self.iblock_of_j[self.cycler_slow.next()]
        self.get_block_proposal(P, current_iblock_slow)

    def get_proposal_fast(self, P):
        current_iblock_fast = self.iblock_of_j[
            self.cycler_slow.n + self.cycler_fast.next()]
        self.get_block_proposal(P, current_iblock_fast)

    def get_block_proposal(self, P, iblock):
        vec_standarised = self.proposer[iblock].propose_vec(self.propose_scale)
        P[self.i_of_j[self.j_start[iblock]:]] += (self.transform[iblock]
                                                  .dot(vec_standarised))

    def set_covariance(self, propose_matrix):
        """
        Take covariance of sampled parameters (propose_matrix), and construct orthonormal
        parameters where orthonormal parameters are grouped in blocks by speed, so changes
        in slowest block changes slow and fast parameters, but changes in the fastest
        block only changes fast parameters

        :param propose_matrix: covariance matrix for the sampled parameters.
        """
        if propose_matrix.shape[0] != self.d():
            log.error("The covariance matrix does not have the correct dimension: "
                      "it's %d, but it should be %d.",
                      propose_matrix.shape[0], self.d())
            raise HandledException
        if not (np.allclose(propose_matrix.T,propose_matrix) and
                np.all(np.linalg.eigvals(propose_matrix) > 0)):
            log.error("The given covmat is not a positive-definite, "
                      "symmetric square matrix.")
            raise HandledException
        self.propose_matrix = propose_matrix.copy()
        propose_matrix_j_sorted = self.propose_matrix[np.ix_(self.i_of_j,self.i_of_j)]
        sigmas_diag = np.diag(np.sqrt(np.diag(propose_matrix_j_sorted)))
        invsigmas_diag = np.linalg.inv(sigmas_diag)
        corr = invsigmas_diag.dot(propose_matrix_j_sorted).dot(invsigmas_diag)
        L = np.linalg.cholesky(corr)
        # Store the basis as transformation matrices
        self.transform = []
        for iblock, bp in enumerate(self.proposer):
            j_start = self.j_start[iblock]
            j_end = j_start + bp.n
            self.transform += [sigmas_diag[j_start:,j_start:]
                               .dot(L[j_start:,j_start:j_end])]

    def get_covariance(self):
        return self.propose_matrix.copy()



# Python 2/3 compatibility
from __future__ import division

# Global
import numpy as np

# Local
from cobaya.likelihood import Likelihood

class _test(Likelihood):
    """
    Likelihood that evaluates to 1.
    """

    def logp(self, **params_values):
        self.wait()
        return 0.



"""
.. module:: sampler

:Synopsis: Prototype sampler class and sampler loader
:Author: Jesus Torrado

cobaya includes by default an
:doc:`advanced Monte Carlo Markov Chain (MCMC) sampler <sampler_mcmc>`
(a direct translation from `CosmoMC <http://cosmologist.info/cosmomc/>`_ and a dummy
:doc:`evaluate <sampler_evaluate>` sampler, that simply evaluates the posterior at a given
(or sampled) reference point. It also includes an interface to the 
:doc:`PolyChord sampler <sampler_polychord>` (needs to be installed separately).

The sampler to use is specified by a `sampler` block in the input file, whose only member
is the sampler used, containing some options, if necessary.

.. code-block:: yaml

   sampler:
     mcmc:
       max_samples: 1000

or 

.. code-block:: yaml

   sampler:
     polychord:
       path: /path/to/cosmo/PolyChord

Samplers can in general be swapped in the input file without needing to modify any other
block of the input.

In the cobaya code tree, each sampler is placed in its own folder, containing a file 
defining the sampler's class, which inherits from the :class:`cobaya.Sampler`, and a 
``defaults.yaml`` file, containing all possible user-specified options for the sampler and
their default value. Whatever option is defined in this file becomes automatically an
attibute of the sampler's instance.

To implement your own sampler, or an interface to an external one, simply create a folder
under the ``cobaya/cobaya/samplers/`` folder and include the two files described above.
Your class needs to inherit from the :class:`cobaya.Sampler` class below, and needs to
implement only the methods ``initialise``, ``run``, ``close``, and ``products``.

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
from importlib import import_module
import numpy as np

# Local
from cobaya.conventions import _sampler, package
from cobaya.tools import get_class
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)


class Sampler(object):
    """Prototype of the sampler class."""

    # What you *must* implement to create your own sampler:
    
    def initialise(self):
        """
        Initialises the sampler: prepares the samples' collection,
        prepares the output, deals with MPI scheduling, imports an external sampler, etc.

        Options defined in the ``defaults.yaml`` file in the sampler's folder are
        automatically recognised as attributes, with the value given in the input file, 
        if redefined there.

        The prior and likelihood are also accesible through the attributes with the same
        names.
        """
        pass

    def run(self):
        """
        Runs the main part of the algorithm of the sampler.
        Normally, it looks somewhat like

        .. code-block:: python
           
           while not [convergence criterion]:
               [do one more step]
               [update the collection of samples]
        """
        pass

    def close(self):
        """
        Finalises the sampler, if something needs to be done
        (e.g. generating additional output).
        """
        pass

    def products(self):
        """
        Returns the products expected in a scripted call of cobaya,
        (e.g. a collection of smaples or a list of them).
        """
        return None
    
    # Private methods: just ignore them:
    def __init__(self, info_sampler, parametrisation, prior, likelihood, output):
        """
        Actual initialisation of the class. Loads the default and input information and
        call the custom ``initialise`` method.
        
        [Do not modify this one.]
        """
        self.name = None
        # Load default and input info
        self.parametrisation = parametrisation
        self.prior = prior
        self.likelihood = likelihood
        self.output = output
        # Load info of the sampler
        for k in info_sampler:
            setattr(self, k, info_sampler[k])
        # Number of times the posterior has been evaluated
        self.n_eval = 0
        self.initialise()

    def logposterior(self, params_values):
        """
        Returns (logposterior,logprior,[loglikelihoods]) for a set of parameter values.
        If passes an empty list through ``derived``,
        it gets populated it with the derived parameters' values.
        """
        logprior = self.prior.logp(params_values)
        logpost = logprior
        logliks = []
        if logprior > -np.inf:
            derived = []
            logliks = self.likelihood.logps(
                self.parametrisation.to_input(params_values), derived=derived)
            logpost += sum(logliks)
            derived_sampler = self.parametrisation.to_derived(derived)
        else:
            derived_sampler = []
        self.n_eval += 1
        return logpost, logprior, logliks, derived_sampler

    # Python magic for the "with" statement
    def __enter__(self):
        return self
    def __exit__(self, exception_type, exception_value, traceback):
        self.close()


def get_Sampler(info_sampler, parametrisation, prior, likelihood, output_file):
    """
    Auxiliary function to retrieve and initialise the requested sampler.
    """
    try:
        name = info_sampler.keys()[0]
    except AttributeError:
        log.error("The sampler block must be a dictionary 'sampler: {options}'.")
        raise HandledException
    sampler_class = get_class(name, kind=_sampler)
    return sampler_class(info_sampler[name], parametrisation, prior, likelihood, output_file)



"""
.. module:: conventions

:Synopsis: Some naming conventions to make the life of the maintainer easier.
:Author: Jesus Torrado

"""
# Package name (for importlib)
# (apparently __package__ is only defined if you import something locally.
package = __name__.rpartition('.')[0]

# Names for block and fields in the input
_likelihood = "likelihood"
_prior      = "prior"
_theory     = "theory"
_sampler    = "sampler"
_params     = "params"
_p_dist     = "dist"
_p_ref      = "ref"
_p_label    = "latex"
_p_drop     = "drop"
_p_derived  = "derived"
_p_proposal = "proposal"
_debug      = "debug"
_debug_file = "debug_file"
_output_prefix = "output_prefix"
_path_install = "path_to_modules"
_external = "external"

# Separator for different names.
# Its manual inclusion in a string anywhere else (e.g. a parameter name) is forbidden.
separator = "__"

# Names for the samples' fields internally and in the output
_weight        = "weight"        # sample weight
_minuslogpost  = "minuslogpost"  # log-posterior, or in general the total log-probability
_minuslogprior = "minuslogprior" # log-prior
_chi2          = "chi2"          # chi^2 = -2 * loglik
_derived_pre   = "derived"+separator  # prefix for derived parameters

# Output files
_input_suffix = "input"
_full_suffix  = "full"

# Internal package structure
_defaults_file = "defaults.yaml"
subfolders = {_likelihood: "likelihoods",
              _sampler: "samplers",
              _theory: "theories"}



"""
.. module:: theories.camb

:Synopsis: Managing the CAMB cosmological code
:Author: Jesus Torrado

.. |br| raw:: html

   <br />

This module imports and manages the CAMB cosmological code.

.. note::

   **If you use this cosmological code, please cite it as:**
   |br|
   `A. Lewis, A. Challinor, A. Lasenby, "Efficient computation of CMB anisotropies in closed FRW"
   (arXiv:astro-ph/9911177) <https://arxiv.org/abs/astro-ph/9911177>`_
   |br|
   `C. Howlett, A. Lewis, A. Hall, A. Challinor, "CMB power spectrum parameter degeneracies in the era of precision cosmology"
   (arXiv:1201.3654) <https://arxiv.org/abs/1201.3654>`_


Usage
-----

If you are using a likelihood that requires some observable from CAMB, simply add CAMB
to the theory block.

You can specify any parameter that CAMB understands within the ``theory``
sub-block of the ``params`` block:

.. code-block:: yaml

   theory:
     camb:

   params:
     theory:
       [any param that CAMB understands, fixed, sampled or derived]


Installation
------------

Pre-requisites
^^^^^^^^^^^^^^

**cobaya** calls CAMB using its Python interface, which requires that you compile CAMB
using the GNU gfortran compiler version 4.9 or later. To check if you fulfil that
requisite, type ``gfortran --version`` in the shell, and the first line should look like

.. code::

   GNU Fortran ([your OS version]) [gfortran version] [release date]

Check that ``[gfortran's version]`` is at least 4.9. If you get an error instead, you need
to install gfortran (contact your local IT service).


Automatic installation
^^^^^^^^^^^^^^^^^^^^^^

If you do not plan to modify CAMB, the easiest way to install it is using the
:doc:`automatic installation script <installation_ext>`. Just make sure that
``theory: camb:`` appears in one of the files passed as arguments to the installation
script.


Manual installation (or using your own version)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you are planning to modify CAMB or use an already modified version,
you should not use the automatic installation script. Use the installation method that
best adapts to your needs:

* [**Recommended**]
  To install an up-to-date CAMB locally and use git to keep track of your changes,
  `fork the CAMB repository in Github <https://github.com/cmbant/CAMB>`_
  (follow `this instructions <https://help.github.com/articles/fork-a-repo/>`_) and clone
  it in some folder of your choice, say ``/path/to/theories/CAMB``:

  .. code:: bash
     
      $ cd /path/to/theories/
      $ git clone https://[YourGithubUser]@github.com/[YourGithubUser]/CAMB.git
      $ cd CAMB/pycamb
      $ python setup.py build
     
* To install an up-to-date CAMB locally, if you don't care about keeping track of your
  changes (and don't want to use ``git``), do:

  .. code:: bash

      $ cd /path/to/theories/
      $ wget https://github.com/cmbant/CAMB/archive/master.zip
      $ unzip master.zip
      $ rm master.zip
      $ mv CAMB-master CAMB       
      $ cd CAMB/pycamb
      $ python setup.py build

* To use your own version, assuming it's placed under ``/path/to/theories/CAMB``,
  just make sure it is compiled (and that the version on top of which you based your
  modifications is old enough to have the ``pycamb`` interface implemented.

In the three cases above, you **must** specify the path to your CAMB installation in
the input block for CAMB (otherwise a system-wide CAMB may be used instead):

.. code:: yaml

   theory:
     camb:
       path: /path/to/theories/CAMB

.. note::

   In any of these methods, you should **not** install CAMB as python package using
   ``python setup.py install --user``, as the official instructions suggest.
   It is actually safer not to do so if you intend to switch between different versions or
   modifications of CAMB.


Modifying CAMB
--------------

If you modify CAMB and add new variables, you don't need to let **cobaya** now,
but make sure
that the variables you create are exposed in its Python interface (contact CAMB's
developers if you need help with that).

.. todo::

   Point somewhere to the CAMB documentation where how to make these modifications
   is explained.

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import sys
import os
from copy import deepcopy
import numpy as np
from collections import namedtuple

# Local
from cobaya.theory import Theory
from cobaya.log import HandledException
from cobaya.install import user_flag_if_needed

# Logger
import logging
log = logging.getLogger(__name__)

# Result collector
collector = namedtuple("collector", ["method", "kwargs"])

class camb(Theory):

    def initialise(self):
        """Importing CAMB from the correct path, if given."""
        if self.path:
            log.info("Importing *local* CAMB from "+self.path)
            if not os.path.exists(self.path):
                log.error("The given folder does not exist: '%s'", self.path)
                raise HandledException
            pycamb_path = os.path.join(self.path, "pycamb")
            if not os.path.exists(pycamb_path):
                log.error("Either CAMB is not in the given folder, '%s', or you are "
                          "using a very old version without the `pycamb` Python interface.",
                          self.path)
                raise HandledException
            sys.path.insert(0, pycamb_path)
        else:
            log.info("Importing *global* CAMB.")
        try:
            import camb
        except ImportError:
            log.error(
                "Couldn't find the CAMB python interface.\n"
                "Make sure that you have compiled it, and that you either\n"
                " (a) specify a path (you didn't) or\n"
                " (b) install the Python interface globally with\n"
                "     '/path/to/camb/pycamb/python setup.py install --user'")
            raise HandledException
        self.camb = camb
        # Generate states, to avoid recomputing
        self.n_states = 3
        self.states = [{"CAMBparams": None, "CAMBresults": None,
                        "params": None, "derived": None, "last": 0}
                       for i in range(self.n_states)]
        # Dict of named tuples to collect requirements and computation methods
        self.collectors = {}
        # Additional input parameters to pass to CAMB
        self.also = {}
        # patch: if cosmomc_theta is used, CAMB needs to be passed explicitly "H0=None"
        if all((p in self.input_params) for p in ["H0", "cosmomc_theta"]):
            log.error("Can't pass both H0 and cosmomc_theta to Camb.")
            raise HandledException
        if "cosmomc_theta" in self.input_params or "cosmomc_theta":
            self.also["H0"] = None

    def current_state(self):
       lasts = [self.states[i]["last"] for i in range(self.n_states)]
       return self.states[lasts.index(max(lasts))]

    def set(self, params_values_dict, i_state):
        # Feed the arguments defining the cosmology to the cosmological code
        # Additionally required input parameters
        args = deepcopy(self.also)
        # Sampled -- save the state for avoiding recomputing later
        args.update(params_values_dict)
        # Precision (fixed at the theory block level)
        if self.precision:
            args.update(self.precision)
        # Generate and save
        log.debug("Setting parameters: %r", args)
        self.states[i_state]["params"] = params_values_dict
        try:
            self.states[i_state]["CAMBparams"] = self.camb.set_params(**args)
        except Exception:
            log.error("Error setting CAMB parameters -- see CAMB's error trace below.\n"
                      "The parameters were %r", args)
            raise

    def compute(self, derived=None, **params_values_dict):
        lasts = [self.states[i]["last"] for i in range(self.n_states)]
        try:
            # are the parameter values there already?
            i_state = (i for i in range(self.n_states)
                       if self.states[i]["params"] == params_values_dict).next()
            # Get (pre-computed) derived parameters
            if derived == {}:
                derived.update(dict([[p,v] for p,v in
                                     zip(self.output_params, self.states[i_state]["derived"])]))
            log.debug("Re-using computed results (state %d)", i_state)
        except StopIteration:
            # update the (first) oldest one and compute
            i_state = lasts.index(min(lasts))
            log.debug("Computing (state %d)", i_state)
            self.set(params_values_dict, i_state)
            # Compute the necessary products
            self.states[i_state]["CAMBresults"] = \
                self.camb.get_results(self.states[i_state]["CAMBparams"])
            # extract all requested products
            for product in self.collectors:
                method = getattr(self.states[i_state]["CAMBresults"],
                                 self.collectors[product].method)
                self.states[i_state][product] = method(
                    self.states[i_state]["CAMBparams"], **self.collectors[product].kwargs)
            # Prepare derived parameters
            if derived == {}:
                derived.update(self.get_derived(i_state))
                # Careful: next step must keep the order
                self.states[i_state]["derived"] = [derived[p] for p in self.output_params]
        # make this one the current one by decreasing the antiquity of the rest
        for i in range(self.n_states):
            self.states[i]["last"] -= max(lasts)
        self.states[i_state]["last"] = 1
        return True

    def needs(self, arguments):
        # Computed quantities required by the likelihood
        for k,v in arguments.items():
            if k == "l_max":
                # Take the max of the requested ones
                self.also["lmax"] = max(v,self.also.get("lmax",0))
            elif k == "Cl":
                self.collectors["powers"] = collector(
                    method="get_cmb_power_spectra",
                    kwargs={"spectra": ["total"], "raw_cl": True})
            else:
                log.error("'%s' does not understand the requirement '%s:%s'.",
                          self.__class__.__name__,k,v)
                raise HandledException
        # Derived parameters (if some need some additional computations)
        # ...

    def get_derived(self, i_state):
        """
        Returns a dictionary of derived parameters with their values, using the
        state #`i_state`.
        """
        derived = {}
        params = self.states[i_state]["CAMBparams"]
        results = self.states[i_state]["CAMBresults"]
        # Create list of *general* functions to get derived parameters
        def from_params(p):
            return getattr(params, p, None)
        def from_std(p):
            if p in self.camb.model.derived_names:
                if not hasattr(self, "get_derived_params"):
                    self.get_derived_params = results.get_derived_params()
            return getattr(self, "get_derived_params", {}).get(p, None)
        def from_getter(p):
            return getattr(params, "get_"+p, lambda: None)()
        # fill the list
        for p in self.output_params:
            for f in [from_params, from_std, from_getter]:
                derived[p] = f(p)
                if derived[p] != None:
                    break
            # specific calls, if general ones above failed:
            if p == "sigma8":
                derived[p] = results.get_sigma8()[0]
            if derived[p] == None:
                log.error("Derived param '%s' not implemented in the CAMB interface", p)
                raise HandledException
        # Cleanup (necessary!)
        try:
            del(self.get_derived_params)
        except AttributeError:
            pass
        return derived

    def get_cl(self):
        """
        Returns the power spectra in microK^2.
        """
        current_state = self.current_state()
        # get C_l^XX from the cosmological code
        cl_camb = deepcopy(current_state["powers"]["total"])
        cl = {"ell": np.arange(cl_camb.shape[0]),
              "tt": cl_camb[:,0], "te": cl_camb[:,3], "ee": cl_camb[:,1], "bb":cl_camb[:,2]}
        # convert dimensionless C_l's to C_l in muK**2
        T = current_state["CAMBparams"].TCMB
        for key in cl.iterkeys():
            if key not in ['pp', 'ell']:
                cl[key][2:] = cl[key][2:] *(T*1.e6)**2
        return cl


# Installation routines ###################################################################

def is_installed(**kwargs):
    try:
        import camb
    except:
        return False
    return True

def install(force=False, **kwargs):
    import pip
    exit_status = pip.main(["install", "camb", "--upgrade"] + user_flag_if_needed() +
                            (["--force-reinstall"] if force else []))
    if exit_status:
        return False
    return True
